# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .abc import *
from ..attribute import *
from tramway.core.xyt import crop
from . import collections as helper
import warnings
import numpy as np
import pandas as pd
from collections.abc import Sequence, Set
from tramway.tessellation.base import Partition
from rwa.lazy import lazytype
from collections import defaultdict
import re


class BaseRegion(AnalyzerNode):
    """
    This class should not be directly instanciated.
    It brings basic functionalities to all the available representations
    of regions of interest, merely labelling and analysis registration.
    """
    __slots__ = ('_spt_data','_label')
    def __init__(self, spt_data, label=None, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._spt_data = spt_data
        self._label = label
    @property
    def label(self):
        if callable(self._label):
            self._label = self._label()
        return self._label
    @property
    def source(self):
        return self._spt_data.source
    def get_sampling(self, label=None, unique=True, _type=Partition):
        if callable(label):
            label = label(self.label)
        elif label is None:
            label = self.label
        if isinstance(label, str):
            labels = [label]
        else:
            labels = list(label)
        analyses = [ self._spt_data.get_sampling(_label) for _label in labels ]
        if _type:
            analyses = [ a for a in analyses if issubclass(lazytype(a._data), _type) ]
        if unique:
            if analyses[1:]:
                raise ValueError('label is not specific enough; multiple samplings match')
            elif not analyses:
                raise KeyError("could not find label '{}'".format(label))
            return analyses[0]
        else:
            return analyses
    def add_sampling(self, sampling, label=None, comment=None):
        if callable(label):
            label = label(self.label)
        elif label is None:
            label = self.label
        return self._spt_data.add_sampling(sampling, label, comment)
    def autosaving(self, *args, **kwargs):
        return self._spt_data.autosaving(*args, **kwargs)
    def discard_static_trajectories(self, df, **kwargs):
        return self._spt_data.discard_static_trajectories(df, **kwargs)
    @property
    def _mpl_impl(self):
        from .mpl import Mpl
        return Mpl
    @property
    def mpl(self):
        return self._mpl_impl(self)

class IndividualROI(BaseRegion):
    """ for typing only """
    __slots__ = ()
    pass

class BoundingBox(IndividualROI):
    """ See :class:`BoundingBoxes`. """
    __slots__ = ('_bounding_box',)
    def __init__(self, bb, label, spt_data, **kwargs):
        IndividualROI.__init__(self, spt_data, label, **kwargs)
        self._bounding_box = bb
    def crop(self, df=None):
        _min,_max = self._bounding_box
        if df is None:
            df = self._spt_data.dataframe
        n_space_cols = len([ col for col in 'xyz' if col in df.columns ])
        if n_space_cols < _min.size:
            assert _min.size == n_space_cols + 1
            df = df[(_min[-1] <= df['t']) & (df['t'] <= _max[-1])]
            _min, _max = _min[:-1], _max[:-1]
        df = crop(df, np.r_[_min, _max-_min], add_deltas='n' in df.columns)
        return df
    @property
    def bounding_box(self):
        return self._bounding_box
    def crop_frames(self, **kwargs):
        """
        Iterates and crops the image frames.

        `kwargs` are passed to the :class:`~tramway.analyzer.images.Images`
        :meth:`~tramway.analyzer.images.Images.crop_frames` method.
        """
        yield from self._spt_data.get_image().crop_frames(self.bounding_box, **kwargs)

class SupportRegion(BaseRegion):
    """
    Union of overlapping ROI.
    """
    __slots__ = ('_sr_index','_all_roi')
    def __init__(self, r, regions, spt_data, **kwargs):
        BaseRegion.__init__(self,
                spt_data,
                r if isinstance(r, str) else regions.regions.region_label(r),
                **kwargs)
        self._sr_index = r
        self._all_roi = regions
    @property
    def _support_regions(self):
        return self._all_roi.regions
    def crop(self, df=None):
        if df is None:
            df = self._spt_data.dataframe
        df = self._support_regions.crop(self._sr_index, df, add_deltas='n' in df.columns)
        return df
    @property
    def bounding_box(self):
        if isinstance(self._support_regions, helper.UnitRegions):
            return self._support_regions[self._sr_index]
        else:
            minima, maxima = zip(*[ self._support_regions.unit_region[u] \
                for u in self._support_regions[self._sr_index] ])
            return np.min(np.stack(minima, axis=0), axis=0), np.max(np.stack(maxima, axis=0), axis=0)
    def crop_frames(self, **kwargs):
        """
        Iterates and crops the image frames, based on :attr:`bounding_box`.

        `kwargs` are passed to the :class:`~tramway.analyzer.images.Images`
        :meth:`~tramway.analyzer.images.Images.crop_frames` method.
        """
        yield from BoundingBox.crop_frames(self, **kwargs)
    def add_metadata(self, sampling, inplace=False):
        """
        Modifies `sampling` or returns a modified copy of `sampling`.

        In both cases, returns the resulting :class:`Partition` object.
        """
        if not inplace:
            import copy
            sampling = copy.copy(sampling)
            sampling.param = copy.copy(sampling.param)
        metadata = dict(
                label=self.label,
                bounding_box=self._all_roi.bounding_boxes(self._sr_index),
                )
        try:
            self._parent.add_metadata(metadata)
        except AttributeError:
            pass
        try:
            sampling.param['roi'].update(metadata)
        except KeyError:
            sampling.param['roi'] = metadata
        return sampling
    def add_sampling(self, sampling, label=None, comment=None, metadata='inplace'):
        """
        If ``metadata=='inplace'`` (default), `sampling` is modified to store
        ROI-related information.

        Pass ``metadata=True`` instead to store a copy of `sampling` in the
        analysis tree.
        Note however that `sampling` is copied only if modified.
        """
        if metadata:
            sampling = self.add_metadata(sampling, metadata == 'inplace')
        super().add_sampling(sampling, label, comment)

class FullRegion(BaseRegion):
    """
    Wraps the full dataset; does not actually crop.

    A `FullRegion` can be both an individual ROI and a support region.
    """
    __slots__ = ()
    def crop(self, df=None):
        if df is None:
            df = self._spt_data.dataframe
        return df
    def crop_frames(self, **kwargs):
        """
        .. note::

            Time cropping is not supported yet.

        """
        if self.time_support is not None:
            self.logger.warning('time cropping is not supported yet')
        yield from self._spt_data.as_frames(**kwargs)


class DecentralizedROIManager(AnalyzerNode):
    """
    This class allows to iterate over the ROI defined at the level of
    each SPT data item.
    """
    __slots__ = ('_records',)
    def __init__(self, first_record=None, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._records = set()
        if first_record is not None:
            self._register_decentralized_roi(first_record)
    def _register_decentralized_roi(self, has_roi):
        if isinstance(has_roi.roi, ROIInitializer):
            self.logger.warning('cannot register an uninitialized ROI attribute')
            return
        self._records.add(has_roi)
        has_roi.roi._global = self
    def _update_decentralized_roi(self, known_record, new_record):
        self._records.remove(known_record)
        self._register_decentralized_roi(new_record)
    def reset(self):
        self._records = set()
    def self_update(self, op):
        raise NotImplementedError('why for?')
        self._parent._roi = op(self)
    def as_individual_roi(self, index=None, collection=None, source=None, **kwargs):
        """ Generator function; loops over all the individual roi.

        Filtering is delegated to the individual *SPTDataItem.roi* attributes.

        A *callable* filter takes a single key (*int* for indices, *str* for labels and paths)
        and returns a *bool*.
        
        Arguments:
            
            index (*int*, *set* of *int*, *sequence* of *int*, or *callable*):
                individual ROI index filter; indices apply within a collection
                
            collection (*str*, *set* of *str*, *sequence* of *str*, or *callable*):
                collection label filter
            
            source (*str*, *set* of *str*, *sequence* of *str*, or *callable*):
                SPT data source filter
            
        """
        if source is None:
            for rec in self._records:
                yield from rec.roi.as_individual_roi(index, collection, **kwargs)
        else:
            if callable(source):
                sfilter = source
            else:
                sfilter = lambda s: s==source
            for rec in self._records:
                if sfilter(rec.source):
                    yield from rec.roi.as_individual_roi(index, collection, **kwargs)
    def as_support_regions(self, index=None, source=None, **kwargs):
        """ Generator function; loops over all the support regions.

        Support regions are equivalent to individual ROI if *group_overlapping_roi*
        was set to :const:`False`.

        Filtering is delegated to the individual SPT data block
        :attr:`~tramway.analyzer.spt_data.SPTDataItem.roi` attributes.

        A *callable* filter takes a single key (*int* for indices, *str* for paths)
        and returns a *bool*.
        
        Arguments:
            
            index (*int*, *set* of *int*, *sequence* of *int*, or *callable*):
                support region index filter
            
            source (*str*, *set* of *str*, *sequence* of *str*, or *callable*):
                SPT data source filter
                
        """
        if source is None:
            for rec in self._records:
                yield from rec.roi.as_support_regions(index, **kwargs)
        else:
            if callable(source):
                sfilter = source
            else:
                sfilter = lambda s: s==source
            for rec in self._records:
                if sfilter(rec.source):
                    yield from rec.roi.as_support_regions(index, **kwargs)
    def __iter__(self):
        if self.group_overlapping_roi is False:
            yield from self.as_support_regions()
        else:
            raise AttributeError(type(self).__name__+' object is not iterable; call methods as_support_regions() or as_individual_roi()')
    @property
    def group_overlapping_roi(self):
        group = None
        for rec in self._records:
            try:
                group_ = rec.roi.group_overlapping_roi
            except AttributeError:
                return None
            else:
                if group is None:
                    group = group_
                elif group:
                    if not group_:
                        return None
                else:
                    if group_:
                        return None
        return group

ROI.register(DecentralizedROIManager)


class ROIInitializer(Initializer):
    """
    Initial value for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute.

    *from_...* methods alters the parent attribute which specializes
    into an initialized :class:`ROI` object.
    """
    __slots__ = ()
    def specialize(self, cls, *args, **kwargs):
        Initializer.specialize(self, cls, *args, **kwargs)
        if self._parent is self._eldest_parent:
            if issubclass(cls, DecentralizedROIManager):
                if cls is not DecentralizedROIManager:
                    # post-register decentralized roi again if necessary
                    spt_data, roi = self._parent.spt_data, self._parent.roi
                    for f in spt_data:
                        if isinstance(f, HasROI) and f.roi.initialized:
                            roi._register_decentralized_roi(f)
            else:
                # replace all individual-SPT-item-level roi initializers by mirrors
                spt_data, roi = self._parent.spt_data, self._parent.roi
                if not spt_data.initialized:
                    raise RuntimeError('cannot define ROI as long as the `spt_data` attribute is not initialized')
                for f in spt_data:
                    assert isinstance(f, HasROI)
                    if f.roi.initialized:
                        raise RuntimeError('ROI already defined at the individual SPT data item level')
                    f.roi._from_common_roi(roi)
    ## initializers
    def _from_common_roi(self, roi):
        # special `specialize`
        spt_data = self._parent
        spt_data._roi = spt_data._bear_child( CommonROI, roi )
    def _register_decentralized_roi(self, roi):
        self.specialize( DecentralizedROIManager, roi )
    def from_bounding_boxes(self, bb, label=None, group_overlapping_roi=False):
        """
        Defines ROI as bounding boxes.

        Arguments:

            bb (sequence): collection of bounding boxes, each bounding boxes being
                a pair of lower and upper bounds (*numpy.ndarray*)

            label (str): unique label for the collection

            group_overlapping_roi (bool): if :const:`False`, :meth:`as_support_regions`
                will behave similarly to :meth:`as_individual_roi`, otherwise support
                regions are unions of overlapping ROI

        See also :class:`BoundingBoxes`.
        """
        self.specialize( BoundingBoxes, bb, label, group_overlapping_roi )
    def from_squares(self, centers, side, label=None, group_overlapping_roi=False):
        """
        Defines spatial ROI as centers for squares/cubes of uniform size.

        Centers should be provided as a sequence of :class:`ndarray`
        or an NxD matrix, with D the number of spatial dimensions.

        See also :meth:`from_bounding_boxes`.
        """
        if isinstance(centers, np.ndarray) and not \
                (centers.shape[1:] and 1<centers.shape[1]):
            raise ValueError('ROI centers are not a NxD matrix')
        bb = [ (center-.5*side, center+.5*side) for center in centers ]
        self.from_bounding_boxes(bb, label, group_overlapping_roi)
    def from_ascii_file(self, filepath, size=None,
            label=None, group_overlapping_roi=False):
        """
        Reads the ROI centers or bounds from a text file.

        .. note::

            This initializer can only be called from the decentralized
            :attr:`~HasROI.roi` attribute of an
            :class:`~tramway.analyzer.spt_data.SPTDataItem` item of the
            :class:`~tramway.analyzer.RWAnalyzer`
            :attr:`~tramway.analyzer.RWAnalyzer.spt_data` main attribute.

        See also :class:`ROIAsciiFile`.
        """
        if isinstance(self._parent, HasROI):
            self.specialize( ROIAsciiFile, filepath, size, label, group_overlapping_roi )
        else:
            raise AttributeError('from_ascii_file called for the main roi attribute')
    def from_ascii_files(self, suffix='roi', extension='.txt', size=None,
            label=None, group_overlapping_roi=False, skip_missing=False):
        """
        Reads the ROI centers or bounds from text files -- one file per item in
        :attr:`~tramway.analyzer.RWAnalyzer.spt_data`.

        The file paths are inferred from the SPT data source files (`source` attribute).
        Files are looked for in the same directories as the corresponding SPT data file,
        and are expected to be named with the same basename, plus a suffix and the *.txt*
        extension.

        A hyphen or underscore character is automatically appended left of the suffix if
        necessary.

        If the `source` attribute of the :class:`~tramway.analyzer.spt_data.SPTDataItem`
        items is not defined, a :class:`ValueError` exception is raised.

        .. note::

            Calling this initializer method from a satellite `roi` attribute
            (nested in an :class:`~tramway.analyzer.spt_data.SPTDataItem` data block)
            is equivalent to calling the same initializer from the
            :class:`~tramway.analyzer.RWAnalyzer`
            :attr:`~tramway.analyzer.RWAnalyzer.roi` main attribute.

        See also :class:`ROIAsciiFiles`.
        """
        if not suffix:
            raise ValueError('undefined suffix')
        if isinstance(suffix, (tuple, list)):
            if not (isinstance(label, (tuple, list)) and len(suffix) == len(label)):
                raise ValueError('not as many labels as suffices')
            extra_suffixes = zip(suffix[1:], label[1:])
            suffix, label = suffix[0], label[0]
        else:
            extra_suffixes = ()
        #
        self._eldest_parent.roi.specialize( ROIAsciiFiles, suffix, extension,
                size, label, group_overlapping_roi, skip_missing )
        #
        for suffix, label in extra_suffixes:
            self._eldest_parent.roi.add_collection(label=label, suffix=suffix)
    def from_dedicated_rwa_record(self, label=None, version=None, **kwargs):
        """
        See also :class:`v1_ROIRecord` and :class:`v2_ROIRecord`.
        """
        if isinstance(self._parent, HasROI):
            if version is None:
                self.logger.info('set version=1 to ensure constant behavior in the future')
                version = 1
            if version == 1:
                self.specialize( v1_ROIRecord, label, **kwargs )
            elif version == 2:
                self.specialize( v2_ROIRecord, label, **kwargs )
            else:
                raise ValueError('version {} not supported'.format(version))
        else:
            self.from_dedicated_rwa_records(label, version, _impl=_impl)
    def from_dedicated_rwa_records(self, label=None, version=None, _impl=None):
        """
        See also :class:`ROIRecords`.
        """
        self._eldest_parent.roi.specialize( ROIRecords, label, version, _impl )
    ## in the case no ROI are defined
    def as_support_regions(self, index=None, source=None, return_index=False):
        """ Generator function; loops over all the support regions.
        
        A :class:`ROIInitializer` attribute object does not define any ROI,
        as a consequence a single :class:`FullRegion` object is yielded."""
        if not null_index(index):
            raise ValueError('no ROI defined; cannot seek for the ith ROI')
        if return_index:
            def bear_child(*args):
                return 0, self._bear_child(*args)
        else:
            bear_child = self._bear_child
        try:
            spt_data = self._parent.spt_data
        except AttributeError:
            # decentralized roi (single source)
            if source is not None:
                warnings.warn('ignoring argument `source`', helper.IgnoredInputWarning)
            spt_data = self._parent
            yield bear_child( FullRegion, spt_data )
        else:
            # roi manager (multiple sources)
            if isinstance(spt_data, Initializer):
                raise RuntimeError('cannot iterate not-initialized SPT data')
            if source is None:
                for d in spt_data:
                    yield bear_child( FullRegion, d )
            else:
                if callable(source):
                    sfilter = source
                else:
                    sfilter = lambda s: s==source
                for d in spt_data:
                    if sfilter(d.source):
                        yield bear_child( FullRegion, d )
    def as_individual_roi(self, index=None, collection=None, source=None, **kwargs):
        """ Generator function; loops over all the individual ROI.
        
        A :class:`ROIInitializer` does not define any ROI,
        as a consequence a single :class:`FullRegion` object is yielded."""
        if collection is not None:
            warnings.warn('ignoring argument `collection`', helper.IgnoredInputWarning)
        return self.as_support_regions(index, source, **kwargs)
    def __iter__(self):
        yield from self.as_support_regions()
        #raise AttributeError(type(self).__name__+' object is not iterable; call methods as_support_regions() or as_individual_roi()')

ROI.register(ROIInitializer)


class CommonROI(AnalyzerNode):
    """
    Mirrors the global :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute.

    The individual *SPTDataItem.roi* attributes become :class:`CommonROI` objects
    as soon as the global :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute is
    specialized, so that :meth:`as_support_regions` and :meth:`as_individual_roi`
    iterators delegate to the global attribute.
    """
    __slots__ = ('_global',)
    def __init__(self, roi, parent=None):
        AnalyzerNode.__init__(self, parent)
        self._global = roi
    def as_support_regions(self, index=None, source=None, return_index=False):
        spt_data = self._parent
        if not spt_data.compatible_source(source):
            warnings.warn('ignoring argument `source`', helper.IgnoredInputWarning)
            return
        yield from self._global.as_support_regions(index, spt_data.source, return_index)
    def self_update(self, op):
        raise RuntimeError('cannot alter a mirror attribute')
    def as_individual_roi(self, index=None, collection=None, source=None, return_index=False):
        spt_data = self._parent
        if not spt_data.compatible_source(source):
            warnings.warn('ignoring argument `source`', helper.IgnoredInputWarning)
            return
        yield from self._global.as_individual_roi(index, collection, spt_data.source, return_index)
    def __iter__(self):
        yield from self._global.as_support_regions()
        #raise AttributeError(type(self).__name__+' object is not iterable; call methods as_support_regions() or as_individual_roi()')
    def to_ascii_file(self, filepath, collection=None, **kwargs):
        """
        Exports the regions of interest to a text file.

        Arguments:

            filepath (str or Path): output filepath.

            collection (str): roi collection label.

        All keyword arguments are passed to the common ROI
        object's `to_ascii_file` method.
        See for example :meth:`BoundingBoxes.to_ascii_file`.
        """
        self._global.to_ascii_file(filepath, collection, **kwargs)


class SpecializedROI(AnalyzerNode):
    """
    Basis for initialized :class:`ROI` classes.
    """
    __slots__ = ('_global','_collections')
    def __init__(self, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._global = None
        self._collections = None
    def self_update(self, op):
        self._parent._roi = op(self)
        if self._global is not None:
            # parent spt_data object should still be registered
            assert self._parent in self._global._records
    def as_support_regions(self, index=None, source=None, return_index=False):
        if return_index:
            def bear_child(cls, r, *args):
                return r, self._bear_child(cls, r, *args)
        else:
            bear_child = self._bear_child
        try:
            spt_data = self._parent.spt_data
        except AttributeError:
            # decentralized roi (single source)
            if source is not None:
                warnings.warn('ignoring argument `source`', helper.IgnoredInputWarning)
            spt_data = self._parent
            for r, _ in indexer(index, self._collections.regions, has_keys=True, return_index=True):
                yield bear_child( SupportRegion, r, self._collections, spt_data )
        else:
            # roi manager (one set of regions, multiple sources)
            if isinstance(spt_data, Initializer):
                raise RuntimeError('cannot iterate not-initialized SPT data')
            if source is None:
                for d in spt_data:
                    for r, _ in indexer(index, self._collections.regions, has_keys=True, return_index=True):
                        yield bear_child( SupportRegion, r, self._collections, d )
            else:
                if callable(source):
                    sfilter = source
                else:
                    sfilter = lambda s: s==source
                for d in spt_data:
                    if sfilter(d.source):
                        for r, _ in indexer(index, self._collections.regions, has_keys=True, return_index=True):
                            yield bear_child( SupportRegion, r, self._collections, d )
    as_support_regions.__doc__ = ROI.as_support_regions.__doc__
    def __iter__(self):
        if self.group_overlapping_roi is False:
            yield from self.as_support_regions()
        else:
            raise AttributeError(type(self).__name__+' object is not iterable; call methods as_support_regions() or as_individual_roi()')
    @property
    def group_overlapping_roi(self):
        try:
            return isinstance(self._collections.regions, helper.helper.GroupedRegions)
        except AttributeError:
            return None
    def get_support_region(self, index, collection=None):
        """
        Returns the :class:`SupportRegion` object corresponding to an individual ROI.

        .. note::

            When using individual ROI indices and the parallelization capability of
            :attr:`~tramway.analyzer.RWAnalyzer.pipeline`,
            beware that :meth:`~ROI.as_individual_roi` is not controlled by the proper
            filters, unlike :meth:`~ROI.as_support_regions`, in all the possible
            implementations.
            :meth:`get_support_region` may still be called on workers assigned to
            processing other regions and consequently raise a :class:`RuntimeError`
            exception.

        """
        sr_index = self._collections.regions.unit_to_region(index, collection)
        sr = single(self.as_support_regions(index=sr_index))
        assert sr._sr_index == sr_index
        return sr


class BoundingBoxes(SpecializedROI):
    """
    Bounding boxes are a list of pairs of NumPy row arrays (1xD).

    The first array specifies the lower bounds, the second array the upper bounds.
    """
    __slots__ = ('_bounding_boxes',)
    def __init__(self, bb, label=None, group_overlapping_roi=False, **kwargs):
        SpecializedROI.__init__(self, **kwargs)
        self._collections = helper.Collections(group_overlapping_roi)
        if label is None:
            label = ''
        if isinstance(bb, pd.DataFrame):
            # new in 0.6
            coord = []
            for c in 'xyzt':
                min_ = c+' min' in bb.columns
                max_ = c+' max' in bb.columns
                if min_ and max_:
                    coord.append(c)
                elif c in 'xy':
                    if min_ or max_ or c in bb.columns:
                        raise ValueError(f"expected column names: '{c} min', '{c} max'")
                    else:
                        raise ValueError("x or y coordinates not found")
            bb = [tuple(np.array([
                    bb.at[r,' '.join((c,b))] \
                        for c in coord]) \
                        for b in ('min','max')) \
                        for r in bb.index]
        self._collections[label] = bb
        self._bounding_boxes = {label: bb}
    @property
    def bounding_boxes(self):
        return self._bounding_boxes
    def __setitem__(self, label, bb):
        if label is None:
            label = ''
        self._collections[label] = bb
        self._bounding_boxes = {label: bb}
    def as_individual_roi(self, index=None, collection=None, source=None, return_index=False):
        if return_index:
            def bear_child(i, *args):
                return i, self._bear_child(*args)
        else:
            def bear_child(i, *args):
                return self._bear_child(*args)
        try:
            spt_data = self._parent.spt_data
        except AttributeError:
            # decentralized roi (single source)
            if source is not None:
                warnings.warn('ignoring argument `source`', helper.IgnoredInputWarning)
            spt_data = self._parent
            for label, bbs in indexer(collection, self.bounding_boxes, has_keys=True,
                    return_index=True):
                for i, bb in indexer(index, bbs, return_index=True):
                    roi_label = self._collections[label].roi_label(i)
                    yield bear_child(i, BoundingBox, bb, roi_label, spt_data )
        else:
            for d in spt_data:
                for label, bbs in indexer(collection, self.bounding_boxes, has_keys=True,
                        return_index=True):
                    for i, bb in indexer(index, bbs, return_index=True):
                        roi_label = self._collections[label].roi_label(i)
                        yield bear_child(i, BoundingBox, bb, roi_label, d )
    as_individual_roi.__doc__ = ROI.as_individual_roi.__doc__
    @property
    def index_format(self):
        """
        *str*: Format of the numeric part of the label
        """
        return self._collections.numeric_format
    @index_format.setter
    def index_format(self, fmt):
        self._collections.numeric_format = fmt
    def set_num_digits(self, n):
        """
        Sets the number of digits in the numeric part of the label.
        """
        if not isinstance(n, int):
            raise TypeError('num_digits is not an int')
        self.index_format = n
    def to_ascii_file(self, filepath, collection=None, columns=None, header=True,
            float_format='%.4f', **kwargs):
        """
        Exports the bounding boxes to text file.

        Arguments:

            filepath (str or Path): output filepath.

            collection (str): roi collection label.

            columns (sequence of *str*): column names; default is:
                ['x', 'y'] for 2D, ['x', 'y', 't'] for 3D
                and ['x', 'y', 'z', 't'] for 4D.

            header (bool): print column names on the first line.

            float_format (str): see also :meth:`pandas.DataFrame.to_csv`.

        Additional keyword arguments are passed to :meth:`pandas.DataFrame.to_csv`.
        """
        if collection is None:
            collection = ''
        example_bound, _ = first(self.bounding_boxes[collection])
        if columns is None:
            columns = {
                    2: list('xy'),
                    3: list('xyt'),
                    4: list('xyzt'),
                }[len(example_bound)]
        bounds = np.stack([ np.r_[bounds] for bounds in self.bounding_boxes[collection] ], axis=0)
        actual_columns = []
        for col in columns:
            actual_columns.append(col+' min')
        for col in columns:
            actual_columns.append(col+' max')
        bounds = pd.DataFrame(bounds, columns=actual_columns)
        for arg in ('sep', 'index'):
            try:
                kwargs.pop(arg)
            except KeyError:
                pass
            else:
                warnings.warn(f'ignoring argument `{arg}`', helper.IgnoredInputWarning)
        bounds.to_csv(str(filepath), sep='\t', index=False, header=header,
                float_format=float_format, **kwargs)
    def add_collection(self, label, bb):
        if label is None:
            label = ''
        if label in self._bounding_boxes:
            raise KeyError(f"collection '{label}' is already defined")
        self._collections[label] = bb
        self._bounding_boxes[label] = bb

ROI.register(BoundingBoxes)


class ROIAsciiFile(BoundingBoxes):
    """
    :class:`ROI` class for the decentralized :attr:`HasROI.roi` attributes
    to be loaded from text files.

    A ROI file contains tab-separated columns with a header line.

    The columns represent either centers or bounds.

    ROI center information is formed by :const:`'x'` and :const:`'y'` columns
    (and optionally :const:`'z'` but not :const:`'t'`).
    
    ROI bounds are defined by two columns for each coordinate, for example
    labelled :const:`'x min'` and :const:`'x max'` for coordinate *x*.
    Time can also be represented this way.

    Note that combining spatial center information and time bounds is allowed,
    i.e. :const:`'x'`, :const:`'y'`, :const:`'t min'` and :const:`'t max'`.

    The center information from a ROI file must be complemented with
    the `size` argument/attribute.
    """
    __slots__ = ('_path', '_size')
    def __init__(self, path, size=None, label=None,
            group_overlapping_roi=False, **kwargs):
        SpecializedROI.__init__(self, **kwargs) # not BoundingBoxes.__init__
        self._collections = helper.Collections(group_overlapping_roi)
        if label is None:
            label = ''
        self._bounding_boxes = {label: None}
        self._path = path
        self._size = size
    @property
    def reified(self):
        """ *bool*: :const:`True` if the files have been loaded """
        return not all([ bb is None for bb in self._bounding_boxes.values() ])
    @property
    def filepath(self):
        """ *str*: Path of the ROI file """
        return self._path
    @property
    def size(self):
        """ *float*: ROI size for space components; apply solely to center-defined ROI """
        return self._size
    @size.setter
    def size(self, sz):
        if self.reified:
            raise AttributeError("file '{}' has already been loaded; cannot modify the ROI size anymore".format(self.filepath.split('/')[-1]))
        else:
            self._size = sz
    @property
    def bounding_boxes(self):
        # this is enough to make `as_individual_roi` properly work
        if not self.reified:
            self.load()
        return self._bounding_boxes
    def as_support_regions(self, *args, **kwargs):
        if not self.reified:
            self.load()
        yield from BoundingBoxes.as_support_regions(self, *args, **kwargs)
    def load(self, label=None, filepath=None):
        if filepath is None:
            filepath = self.filepath
        roi = pd.read_csv(filepath, sep='\t')
        center_cols = [ col for col in 'xyzt' if col in roi.columns ]
        bound_cols = [ col for col in 'xyzt' if col+' min' in roi.columns and col+' max' in roi.columns ]
        if 't' in center_cols:
            raise ValueError('cannot interprete a ROI center time')
        if bound_cols:
            if center_cols:
                if 't' in bound_cols and not any([ col in bound_cols for col in 'xyz' ]):
                    # only case allowed
                    coords = [ col in center_cols for col in 'xyz' ]
                    if coords:
                        lower_bounds = (roi[coords] - .5 * self.size).join(roi[['t min']]).values
                        upper_bounds = (roi[coords] + .5 * self.size).join(roi[['t max']]).values
                    else:
                        lower_bounds = roi[['t min']].values
                        upper_bounds = roi[['t max']].values
                        if self._size is not None:
                            self.logger.debug('ROI size is defined but no spatial coordinates were found')
                else:
                    raise ValueError('center and bound information both available at the same time')
            else:
                coords = [ col in bound_cols for col in 'xyz' ]
                # ordering from bound_cols matters
                lower_bounds = roi[[ col+' min' for col in bound_cols ]].values
                upper_bounds = roi[[ col+' max' for col in bound_cols ]].values
                if self._size is not None:
                    self.logger.debug('ROI size does not apply to bounds-defined regions')
        else:
            coords = center_cols
            lower_bounds = roi[coords].values - .5 * self.size
            upper_bounds = roi[coords].values + .5 * self.size
        # last check
        if 'x' in coords:
            if 'y' not in coords:
                raise ValueError('x coordinate found but could not find y coordinate')
        elif 'y' in coords:
            raise ValueError('y coordinate found but could not find x coordinate')
        #
        bounds = list(zip(lower_bounds, upper_bounds))
        #
        if label is None:
            for label in self._bounding_boxes:
                if self._bounding_boxes[label] is None:
                    self._collections[label] = bounds
                    self._bounding_boxes[label] = bounds
                    #break?
        else:
            self._collections[label] = bounds
            self._bounding_boxes[label] = bounds
    def add_collection(self, label, filepath):
        if not self.reified:
            self.load()
        if label is None:
            label = ''
        if label in self._bounding_boxes:
            raise KeyError("collection '{}' is already defined".format(label))
        self.load(label, filepath)

ROI.register(ROIAsciiFile)


class ROIAsciiFiles(DecentralizedROIManager):
    """
    :class:`ROI` class for multiple ROI text files.

    The filepaths are inferred from the :attr:`~..spt_data.SPTDataItem.source`
    attribute of each :class:`~..spt_data.SPTDataItem` in the main
    :attr:`~tramway.analyzer.RWAnalyzer.spt_data` attribute.
    The SPT file extensions are replaced by a suffix (usually *-roi*) plus
    the *.txt* extension.

    See also :class:`ROIAsciiFile` for more information on the format.
    """
    __slots__ = ('_suffix',)
    def __init__(self, suffix='roi', extension='.txt', side=None, label=None,
            group_overlapping_roi=False, skip_missing=False, **kwargs):
        DecentralizedROIManager.__init__(self, **kwargs)
        if not isinstance(suffix, str):
            raise TypeError('suffix is not an str')
        self._suffix = suffix+extension
        # TODO: initialize the decentralized ROI so that _list_files is triggered later
        self._list_files(side, label, group_overlapping_roi, skip_missing)
    def _list_files(self, *args):
        args, skip_missing = args[:-1], args[-1]
        import os.path
        first = True
        for f in self._parent.spt_data:
            if f.source is None or not os.path.isfile(f.source):
                self.logger.warning('cannot identify or find SPT data source: {}'.format(f.source))
                continue
            filepath, _ = os.path.splitext(f.source)
            if first:
                found = False
                for join_with in ('', '-', '_'):
                    candidate_filepath = join_with.join((filepath, self._suffix))
                    if os.path.isfile(os.path.expanduser(candidate_filepath)):
                        found = True
                        filepath = candidate_filepath
                        self._suffix = join_with + self._suffix
                        break
                if not found:
                    if skip_missing:
                        self.logger.info('skipping roi file for source: '+str(f.source))
                        continue
                    raise FileNotFoundError('{}{}{}'.format(
                            os.path.basename(filepath),
                            '' if self._suffix[0] in '-_' else '[-_]',
                            self._suffix,
                        ))
                first = False
            else:
                filepath = filepath + self._suffix
                if not os.path.isfile(os.path.expanduser(filepath)):
                    if skip_missing:
                        self.logger.info('skipping roi file for source: '+str(f.source))
                        continue
                    raise FileNotFoundError('{}{}{}'.format(
                            os.path.basename(filepath),
                            '' if self._suffix[0] in '-_' else '[-_]',
                            self._suffix,
                        ))

            #
            f.roi.from_ascii_file(filepath, *args)
        #
        if first:
            raise FileNotFoundError('not any roi file found')
    def add_collection(self, label, suffix, skip_missing=False):
        import os.path
        first = True
        for f in self._parent.spt_data:
            _, extension = os.path.splitext(f.roi.filepath)
            _suffix = suffix + extension
            filepath, _ = os.path.splitext(f.source)
            if first:
                found = False
                for join_with in ('', '-', '_'):
                    candidate_filepath = join_with.join((filepath, _suffix))
                    if os.path.isfile(os.path.expanduser(candidate_filepath)):
                        found = True
                        _filepath, filepath = filepath, candidate_filepath
                        _suffix = join_with + _suffix
                        break
                if not found:
                    if skip_missing:
                        self.logger.info('skipping roi file for source: '+str(f.source))
                        continue
                    raise FileNotFoundError('{}{}{}'.format(
                            os.path.basename(_filepath),
                            '' if _suffix[0] in '-_' else '[-_]',
                            _suffix,
                        ))
                first = False
            else:
                _suffix = join_with + _suffix
                _filepath, filepath = filepath, filepath + _suffix
                if not os.path.isfile(os.path.expanduser(filepath)):
                    if skip_missing:
                        self.logger.info('skipping roi file for source: '+str(f.source))
                        continue
                    raise FileNotFoundError('{}{}{}'.format(
                            os.path.basename(_filepath),
                            '' if _suffix[0] in '-_' else '[-_]',
                            _suffix,
                        ))

            #
            f.roi.add_collection(label, filepath)
        #
        if first:
            raise FileNotFoundError('not any roi file found')

ROI.register(ROIAsciiFiles)


class v1_ROIRecord(BoundingBoxes):
    """
    :class:`ROI` class for the individual special partitions in an analysis tree.

    This storing strategy is likely to be marked as deprecated.
    """
    __slots__ = ('_helper',)
    def __init__(self, label=None, _impl=None, **kwargs):
        SpecializedROI.__init__(self, **kwargs) # not BoundingBoxes.__init__
        if _impl is None:
            _impl = helper.helper.RoiHelper
        analyses = self._parent.analyses
        self._helper = _impl(analyses.analyses.statefree(), autosave=False, verbose=False)
        if label is None:
            label = ''
        self._bounding_boxes = {label: None}
    @property
    def _collections(self):
        return self._helper.collections
    @_collections.setter
    def _collections(self, roi):
        if roi is not None:
            raise AttributeError("can't set attribute")
    @property
    def reified(self):
        """ *bool*: :const:`True` if the files have been loaded """
        return not all([ bb is None for bb in self._bounding_boxes.values() ])
    @property
    def bounding_boxes(self):
        # this is enough to make `as_individual_roi` properly work
        if not self.reified:
            self.load()
        return self._bounding_boxes
    def as_support_regions(self, *args, **kwargs):
        if not self.reified:
            self.load()
        yield from BoundingBoxes.as_support_regions(self, *args, **kwargs)
    def load(self):
        for coll_label, meta_label in self._helper.get_meta_labels().items():
            self._bounding_boxes[coll_label] = \
                    self._helper.collections[coll_label] = \
                            self._helper.get_bounding_boxes(meta_label=meta_label)


class ROIRecoveredFromSampling(BoundingBoxes):
    """
    :class:`ROI` class that recovers bounding-box information from
    the sampling (:class:`Partition` records) found in the analysis tree,
    provided that the records are labelled the default way, *i.e.*
    with no label or a static label prefix.

    The bounding boxes are inferred from the data and, if a sliding
    time window was defined, from the windowing start time.
    If the SPT data are location or trajectory data, they are converted
    into translocation data prior to determining the spatial bounding
    box.

    The resulting bounding boxes may be tighter than the original bounding
    boxes, but the sampling should not be affected.
    This may work only if the original ROI were defined with
    `group_overlapping_roi=False` for the sampling step.

    A major drawback of this approach lies in the fact that ROI centers
    cannot be recovered, as they are usually implicitly encoded in the
    bounds (middle point).

    A word of warning, though: the procedure of determining the bounding
    boxes loads all the data from the *.rwa* files, if the analysis
    trees are to be loaded from files.

    As the name hints, this class was designed for recovering ROI
    information in the cases this information is no longer available
    elsewhere.
    """
    __slots__ = ()
    def __init__(self, label=None, group_overlapping_roi=False, **kwargs):
        # check input args
        if label is not None:
            raise NotImplementedError('cannot load a single ROI collection')
        if group_overlapping_roi:
            raise NotImplementedError('ROI grouping is not supported')
        # init
        SpecializedROI.__init__(self, **kwargs) # not BoundingBoxes.__init__
        self._collections = helper.Collections(group_overlapping_roi=False)
        if label is None:
            label = ''
        self._bounding_boxes = {label: None}
    @property
    def reified(self):
        """ *bool*: :const:`True` if the files have been loaded """
        return not all([ bb is None for bb in self._bounding_boxes.values() ])
    @property
    def bounding_boxes(self):
        # this is enough to make `as_individual_roi` properly work
        if not self.reified:
            self.load()
        return self._bounding_boxes
    def as_support_regions(self, *args, **kwargs):
        if not self.reified:
            self.load()
        yield from BoundingBoxes.as_support_regions(self, *args, **kwargs)
    def split_roi_label(self, label):
        pattern = r'(?P<collection>.*[^0-9])(?P<index>[0-9]+(-[0-9]+)*)'
        m = re.fullmatch(pattern, label)
        if m:
            prefix = m.group('collection').rstrip()
            indices = m.group('index').split('-')
            return prefix, indices
        else:
            raise KeyError('not all the artefacts below the root SPT dataframe are labelled as ROI')
    def infer_bounding_box(self, sampling):
        bb = sampling.bounding_box
        assert 'x' in bb.columns
        assert 'y' in bb.columns
        assert 't' in bb.columns
        cols = [ col for col in 'xyzt' if col in bb.columns ]
        try:
            time_segments = sampling.tessellation.time_lattice
        except AttributeError:
            pass
        else:
            bb = bb.copy()
            bb.at['min','t'] = time_segments[0,0]
            bb.at['max','t'] = time_segments[-1,1]
        return (bb[cols].loc['min'].values, bb[cols].loc['max'].values)
    def load(self):
        analyses = self._parent.analyses
        labels = list(analyses.labels)
        collections = defaultdict(dict)
        ndigits = None
        for label in labels:
            collection_label, indices = self.split_roi_label(label)
            if indices[1:]:
                raise NotImplementedError('ROI were defined with group_overlapping_roi=True')
            if ndigits is None:
                ndigits = len(indices[0])
                self.set_num_digits(ndigits)
            index = int(indices[0])
            bounding_box = self.infer_bounding_box(analyses[label].data)
            collections[collection_label][index] = bounding_box
        self._bounding_boxes = {}
        for label in collections:
            imax = max(collections[label].keys())
            ordered_collection, example_bounding_box = [], None
            for i in range(imax+1):
                try:
                    bb = collections[label][i]
                except KeyError:
                    if example_bounding_box is None:
                        raise NotImplementedError('some ROI are missing in the analysis tree') from None
                    bb = (np.zeros_like(example_bounding_box[0]), np.zeros_like(example_bounding_box[1]))
                    self.logger.warning('missing ROI: {}{}'.format(
                            '' if label=='roi' else label+' ', i))
                else:
                    if example_bounding_box is None:
                        example_bounding_box = bb
                ordered_collection.append(bb)
            #
            if label == 'roi':
                label = ''
            self._collections[label] = ordered_collection
            self._bounding_boxes[label] = ordered_collection
        if not self._bounding_boxes:
            raise RuntimeError('no ROI collections found')

v2_ROIRecord = ROIRecoveredFromSampling


class ROIRecords(DecentralizedROIManager):
    """
    :class:`ROI` class for the main :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute.

    See also the corresponding item class:

    * ``version=1``: :class:`v1_ROIRecord` for analysis trees with a meta-partition
        for each ROI collection;
    * ``version=2``: :class:`v2_ROIRecord` to recover ROI definition from the
        the sampling/partition records available in the analysis trees without
        any extra information (no meta-partition).

    """
    __slots__ = ()
    def __init__(self, label=None, version=None, _impl=None, **kwargs):
        DecentralizedROIManager.__init__(self, **kwargs)
        if version is None:
            self.logger.info('set version=1 to ensure constant behavior in the future')
            version = 1
        if version == 1:
            _kwargs = dict(_impl=_impl)
        else:
            _kwargs = {}
        for f in self._parent.spt_data:
            f.roi.from_dedicated_rwa_record(label, version, **_kwargs)


class HasROI(AnalyzerNode):
    """ Class to be inherited from by SPT data item classes.
    
    Maintains a self-modifying :attr:`roi` attribute."""
    __slots__ = ('_roi',)
    def _get_roi(self):
        """
        *ROI*: Regions of interest for the parent data block
        """
        return self._roi
    def _set_roi(self, roi):
        self._roi = roi
        global_roi_attr = self._eldest_parent.roi
        if global_roi_attr.initialized:
            assert isinstance(global_roi_attr, DecentralizedROIManager)
        global_roi_attr._register_decentralized_roi(self)
    roi = selfinitializing_property('roi', _get_roi, _set_roi, ROI)

    def __init__(self, roi=ROIInitializer, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._roi = roi(self._set_roi, parent=self)
    def compatible_source(self, source):
        """
        Returns :const:`True` if filter *source* matches with `self.source`.

        .. note::

            Does not check against the alias.

        """
        if source is None:
            return True
        elif callable(source):
            return source(spt_data.source)
        elif isinstance(source, str):
            return source == spt_data.source
        elif isinstance(source, (Set, Sequence)):
            return spt_data.source in source
        else:
            raise NotImplementedError


__all__ = [ 'ROI', 'ROIInitializer', 'SpecializedROI', 'BoundingBoxes', 'DecentralizedROIManager',
        'BaseRegion', 'FullRegion', 'IndividualROI', 'BoundingBox', 'SupportRegion',
        'CommonROI', 'HasROI', 'ROIAsciiFile', 'ROIAsciiFiles', 'v1_ROIRecord', 'ROIRecords',
        'ROIRecoveredFromSampling', 'v2_ROIRecord']




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
As of version 0.4.7, this package is frozen and is only affected by minor fixes.
"""

from .tessellation import *
from .inference import *
#from .feature import *




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numpy as np
import pandas as pd
from tramway.feature.adjacency import *
from tramway.plot.mesh import *
from tramway.plot.map import *
from tramway.helper.inference import _clip
import traceback
import matplotlib.pyplot as plt


class ContourEditor(object):
    def __init__(self, figure, renderer=None):
        self.figure = figure
        if renderer is None:
            renderer = plt
        self.renderer = renderer
        self._variables = None
        self._cells = None
        self._map = None
        self._cell_ptr = []
        self._area_ptr = []
        self._contour_ptr = []
        self._delaunay_ptr = []
        self._cell_centers = None
        self._cell_adjacency = None
        self._dilation_adjacency = None
        self._delaunay = False
        self._cell = None
        self._step = 1
        self._variable = None
        self._area = True
        self.clip = 4.
        self.debug = False
        self._areas = None
        self.callback = None

        self.cell_marker = 's'
        self.cell_marker_color = 'r'
        self.cell_marker_size = 10
        self.delaunay_color = (.1, .1, .1)

    @property
    def map(self):
        return self._map

    @map.setter
    def map(self, m):
        self._map = m
        if self._map is not None:
            self._variables = None
        self._clear()

    @property
    def variables(self):
        if self._variables is None:
            if self.map is None:
                return []
            self._variables = []
            for v in self.map.columns:
                self._variables.append(v[::-1].split(None, 1)[-1][::-1])
            self._variables = list(set(self._variables))
            self._variables = { v: [ c for c in self.map.columns if c.startswith(v) ] \
                    for v in self._variables }
        return list(self._variables.keys())

    @property
    def cell_centers(self):
        if self._cell_centers is None:
            self._cell_centers = self.cells.tessellation.cell_centers
        return self._cell_centers

    @property
    def cell_adjacency(self):
        if self._cell_adjacency is None:
            if self.cells.tessellation.cell_label is None:
                label = self.cells.location_count
            else:
                label = np.logical_and(0 < self.cells.tessellation.cell_label,
                    0 < self.cells.location_count)
            self._cell_adjacency = self.cells.tessellation.simplified_adjacency(
                label=label, format='csr')
        return self._cell_adjacency

    @property
    def cell_count(self):
        return self.cells.tessellation.cell_adjacency.shape[0]

    @property
    def cells(self):
        return self._cells

    @cells.setter
    def cells(self, cs):
        self._cells = cs
        self._cell_centers = None
        self._cell_adjacency = None
        self._dilation_adjacency = None
        self.map = None
        self._areas = None

    @property
    def dilation_adjacency(self):
        if self._dilation_adjacency is None:
            try:
                self._dilation_adjacency = self.cells.tessellation.diagonal_adjacency
            except AttributeError:
                self._dilation_adjacency = False
            else:
                if self.cells.tessellation.cell_label is None:
                    label = self.cells.location_count
                else:
                    label = np.logical_and(0 < self.cells.tessellation.cell_label,
                        0 < self.cells.location_count)
                self.cells.tessellation.simplified_adjacency(
                    adjacency=self._dilation_adjacency, label=label, format='csr')
        if self._dilation_adjacency is False:
            return self.cell_adjacency
        else:
            return self._dilation_adjacency

    def _clear(self):
        self._cell_ptr = []
        self._area_ptr = []
        self._contour_ptr = []
        self._delaunay_ptr = []
        #self.axes.clear()
        self.figure.clear()

    @property
    def cell(self):
        return self._cell

    @cell.setter
    def cell(self, c):
        if c is None or (-1 <= c and (self.cells is None or c < self.cell_count)):
            self._cell = c
            self.refresh()
        else:
            raise ValueError('cell index not in range (-1,%s)', self.cell_count-1)

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, s):
        if s < 1:
            raise ValueError
        else:
            self._step = s
            self.refresh()

    @property
    def variable(self):
        return self._variable

    @variable.setter
    def variable(self, v):
        if self.cells is None or self.map is None:
            return
        self._clear()
        self._variable = v
        if self.debug or self._delaunay:
            kwargs = {'delaunay': True,
                'color': [self.delaunay_color] * 10,
                'linewidth': .3}
        else:
            kwargs = {'linewidth': 0}
        if self._variables[v][1:]:
            obj = field_map_2d(self.cells, _clip(self.map[self._variables[v]], self.clip),
                figure=self.figure, **kwargs)
        else:
            obj = scalar_map_2d(self.cells, _clip(self.map[v], self.clip),
                figure=self.figure, **kwargs)
        if obj:
            self._delaunay_ptr = obj
            self._delaunay = True
        #self.delaunay = self._delaunay
        self.refresh()

    @property
    def delaunay(self):
        return self._delaunay

    @delaunay.setter
    def delaunay(self, state):
        #if self.debug:
        #       return
        toggle = self._delaunay != bool(state)
        if toggle:
            self._delaunay = not self._delaunay
        if self.cells is None:
            return
        if self._delaunay_ptr:
            if toggle:
                for p in self._delaunay_ptr:
                    p.set_visible(self._delaunay)
                self.draw()
        elif self._delaunay:
            self._delaunay_ptr = plot_delaunay(self.cells, centroid_style=None,
                axes=self.axes)
            self.draw()

    def refresh(self):
        c = self.cell
        if c is None or c < 0 or self.cells is None:
            for p in self._cell_ptr:
                self.axes.lines.remove(p)
            self._cell_ptr = []
            s = 0
        else:
            x, y = self.cell_centers[c]
            if self._cell_ptr:
                self._cell_ptr[0].set_xdata(x)
                self._cell_ptr[0].set_ydata(y)
            else:
                self._cell_ptr = self.axes.plot(x, y, self.cell_marker,
                    markerfacecolor=self.cell_marker_color,
                    markersize=self.cell_marker_size)
            s = self.step
        if 0 < s and self.cells is not None:
            try:
                cs, ok = self.cells.tessellation.contour(c, s, fallback=True,
                        adjacency=self.cell_adjacency,
                        debug=self.debug, cells=self.cells)
            except:
                if self.debug:
                    print(traceback.format_exc())
                cs = []
        else:
            cs = []
        w = None
        if (isinstance(cs, np.ndarray) and cs.size) or cs:
            ws = set()
            for _ws in dilation(c, self.dilation_adjacency, (1, s), cs).values():
                ws |= _ws
            if ws:
                if self.area:
                    w = self.cell_centers[list(ws)]
                    if self._area_ptr:
                        self._area_ptr[0].set_xdata(w[:,0])
                        self._area_ptr[0].set_ydata(w[:,1])
                    else:
                        self._area_ptr = self.axes.plot(w[:,0], w[:,1], 'g.',
                            markersize=6)
            color = 'k' if ok else 'r'
            markeredgecolor = 'y' if ok else 'k'
            z = self.cell_centers[cs]
            z = np.vstack((z, z[0]))
            if self._contour_ptr:
                self._contour_ptr[0].set_xdata(z[:,0])
                self._contour_ptr[0].set_ydata(z[:,1])
                self._contour_ptr[0].set_color(color)
                self._contour_ptr[0].set_markerfacecolor(color)
                self._contour_ptr[0].set_markeredgecolor(markeredgecolor)
            else:
                self._contour_ptr = self.axes.plot(z[:,0], z[:,1], color+'s-',
                    markeredgecolor=markeredgecolor, markerfacecolor=color,
                    markersize=6, linewidth=3)
            if self.callback:
                ws.add(self.cell)
                try:
                    self.callback(self.integral(cs, ws))
                except:
                    print(traceback.format_exc())
        elif self._contour_ptr:
            for p in self._contour_ptr:
                self.axes.lines.remove(p)
            self._contour_ptr = []
        if w is None and self.area:
            for p in self._area_ptr:
                self.axes.lines.remove(p)
            self._area_ptr = []
        self.draw()

    @property
    def axes(self):
        return self.figure.gca()

    @property
    def area(self):
        return self._area

    def draw(self):
        self.renderer.draw()

    @area.setter
    def area(self, a):
        self._area = a
        self.refresh()

    def find_cell(self, pt):
        pt = np.asarray(pt)[np.newaxis,:]
        if isinstance(self.cells.points, pd.DataFrame):
            coord_names = self.cells.points.columns[1:-1]
            pt = pd.DataFrame(data=pt, columns=coord_names)
        elif isstructured(self.cells.points):
            raise NotImplementedError
        ix = self.cells.tessellation.cell_index(pt)
        return ix[0]

    def field(self, cs):
        return self.map.loc[cs, self._variables[self._variable]].values

    def integral(self, contour, inner):
        field = self.field(contour)
        tangent = self.tangent(contour)
        area = self.surface_area(contour, inner)
        if 1 < field.shape[1]:
            return self.curl(field, tangent, area)
        else:
            return self.sum(field, tangent, area)

    def tangent(self, cs):
        X = self.cells.tessellation.cell_centers
        uvw = zip([cs[-1]]+cs[:-1], cs, cs[1:]+[cs[0]])
        return np.vstack([ (X[w]-X[u])*.5 for u, v, w in uvw ])

    def surface_area(self, contour, inner):
        if self._areas is None:
            ncells = self.cells.tessellation.cell_adjacency.shape[0]
            centers = self.cells.tessellation.cell_centers
            vertices = self.cells.tessellation.cell_vertices
            adjacency = self.cells.tessellation.vertex_adjacency.tocsr()
            vert_coords = self.cells.tessellation.vertices
            self._areas = np.zeros(ncells)
            for c in range(ncells):
                u = centers[c]
                verts = set(vertices[c].tolist())
                ordered_verts = []
                next_verts = set(verts) # copy
                try:
                    while verts:
                        vert = next_verts.pop()
                        ordered_verts.append(vert)
                        verts.remove(vert)
                        next_verts = set(adjacency.indices[adjacency.indptr[vert]:adjacency.indptr[vert+1]]) & verts
                except KeyError:
                    if self.debug:
                        print(traceback.format_exc())
                    continue
                for a, b in zip(ordered_verts, ordered_verts[1:]+[ordered_verts[0]]):
                    v, w = vert_coords[a], vert_coords[b]
                    self._areas[c] += abs((v[0]-u[0])*(w[1]-u[1])-(w[0]-u[0])*(v[1]-u[1]))
            self._areas *= .5
        return np.sum(self._areas[list(inner)])

    def sum(self, x, *args):
        return np.mean(x)

    def curl(self, F, dr, A):
        return np.sum(F * dr) / A




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import tramway.inference.unsmooth_ddrift as degraded_dd
import tramway.inference.standard_ddrift as standard_dd


setup = dict(standard_dd.setup) # copy
setup['name'] = ('dd', 'ddrift')
del setup['provides']
setup['infer'] = 'infer_DD'


def infer_DD(cells, diffusivity_prior=None, drift_prior=None, jeffreys_prior=False,
        min_diffusivity=None, max_iter=None, epsilon=None, rgrad=None, **kwargs):

    if diffusivity_prior is None and drift_prior is None:
        return degraded_dd.infer_DD(cells, jeffreys_prior=jeffreys_prior,
                min_diffusivity=min_diffusivity, **kwargs)
    else:
        return standard_dd.infer_smooth_DD(cells, diffusivity_prior, drift_prior, jeffreys_prior,
                min_diffusivity, max_iter, epsilon, rgrad, **kwargs)




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from ..artefact import analysis
from .abc import *
from tramway.helper.tessellation import Tessellate
from copy import deepcopy
from warnings import warn
from .post import TessellationPostProcessingInitializer


def proxy_property(propname, level='default', doc=None):
    """
    Property factory function similar to builtin *property*,
    dedicated to the :class:`TessellerProxy` class.

    For :const:`'tessellate'` level properties, the default level is safe.

    For :const:`'__init__'` level properties, the level must be specified.

    For standard attributes, it is safer to make the level explicit,
    so that some conflicts may be detected earlier (e.g. at module loading).
    """
    if level == 'tessellate':
        def _get(obj):
            return obj._tessellate_kwargs[propname]
        def _set(obj, val):
            assert propname in obj._tessellate_kwargs
            obj._tessellate_kwargs[propname] = val
            obj._explicit_kwargs[propname] = val
    elif level == '__init__':
        def _get(obj):
            return obj._init_kwargs.get(propname, None)
        def _set(obj, val):
            assert propname not in obj._tessellate_kwargs
            obj._init_kwargs[propname] = val
            obj._explicit_kwargs[propname] = val
            obj.reset()
    elif level == 'attr':
        def _get(obj):
            return getattr(obj.tesseller, propname)
        def _set(obj, val):
            setattr(obj.tesseller, propname, val)
            obj._explicit_kwargs[propname] = val
    else:
        def _get(obj):
            return TessellerProxy.__getattr__(obj, propname)
        def _set(obj, val):
            TessellerProxy.__setattr__(obj, propname, val)
    return property(_get, _set, doc=doc)


class TessellerProxy(AnalyzerNode):
    """
    Encapsulates a tessellation plugin as defined in the :mod:`tramway.tessellation` package.

    Attributes are managed with the following rules:

    * if an attribute is set using a proxy property,
      the value is flagged as :const:`'explicit'` and overrides any default value;
    * the arguments in :attr:`_init_kwargs` are passed to the wrapped tesseller's :meth:`__init__`;
      these arguments are intended to prevent :meth:`__init__` from crashing if the later requires
      some arguments to be defined; non-explicit values are likely to be overriden and
      both explicit and non-explicit values may be altered by :meth:`__init__`;
    * the arguments in :attr:`_tessellate_kwargs` are passed to the wrapped tesseller's
      :meth:`tessellate` method; all available arguments should be defined in the proxy's
      :meth:`__init__`, with default values;
    * the wrapped tesseller's attributes can be accessed using proxy properties;
      as a consequence, any key in :attr:`_explicit_kwargs`, that is not in :attr:`_init_kwargs`
      or :attr:`_tessellate_kwargs`, is considered as an actual attribute;
    * an argument should not be defined both in :attr:`_init_kwargs` and :attr:`_tessellate_kwargs`;
    * explicit :meth:`__init__` arguments take precedence over standard attributes;

    """
    __slots__ = ('_tesseller', '_init_kwargs', '_tessellate_kwargs', '_explicit_kwargs', 'alg_name',
            '_post_processing')
    def __init__(self, cls, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._tesseller = cls
        self._reset_kwargs()
        self._explicit_kwargs = {}
        self.alg_name = None
        # self-initializing properties
        self._post_processing = None
        self.post_processing  = TessellationPostProcessingInitializer
    def _reset_kwargs(self):
        self._init_kwargs = dict(ref_distance=None)
        self._tessellate_kwargs = {}
    @property
    def cls(self):
        if isinstance(self._tesseller, type):
            return self._tesseller
        elif isinstance(self._tesseller, tuple):
            return type(self._tesseller[0])
        else:
            return type(self._tesseller)
    def _get_tesseller(self):
        if isinstance(self._tesseller, type):
            cls = self._tesseller
            self._tesseller = cls(**self._init_kwargs)
            for attr in self._iter_explicit_attributes():
                setattr(self._tesseller, attr, self._explicit_kwargs[attr])
        if isinstance(self._tesseller, tuple):
            return self._tesseller[0]
        else:
            return self._tesseller
    tesseller = property(_get_tesseller)
    @property
    def reified(self):
        return isinstance(self._tesseller, tuple)
    def reset(self, clear_explicit_attrs=False):
        """ resets the tesseller to let the calibration parameters vary.
        """
        if not isinstance(self._tesseller, type):
            self._tesseller = self.cls
        self._reset_kwargs()
        if clear_explicit_attrs:
            self._explicit_kwargs = {}
        else:
            for k in self._explicit_kwargs:
                if k in self._tessellate_kwargs:
                    self._tessellate_kwargs[k] = self._explicit_kwargs[k]
                if k in self._init_kwargs:
                    self._init_kwargs[k] = self._explicit_kwargs[k]
    def _iter_explicit_attributes(self):
        for attr in self._explicit_kwargs:
            if not (attr in self._init_kwargs or attr in self._tessellate_kwargs):
                yield attr
    @property
    def helper_cls(self):
        """
            type:
                Tessellation helper class, inheriting from
                :class:`tramway.helper.tessellation.Tessellate`
        """
        return Tessellate
    def calibrate(self, spt_dataframe):
        if not isinstance(self._tesseller, type):
            self.reset()
        helper = self.helper_cls()
        helper.prepare_data(spt_dataframe)
        #
        kwargs = dict(self._init_kwargs)
        kwargs.update(self._tessellate_kwargs)
        helper.setup = dict(make_arguments=kwargs)
        #
        kwargs = helper.standard_parameters(self._init_kwargs.pop('ref_distance', None),
                rel_min_distance=self._init_kwargs.pop('rel_min_distance', None),
                rel_avg_distance=self._init_kwargs.pop('rel_avg_distance', None),
                rel_max_distance=self._init_kwargs.pop('rel_max_distance', None),
                min_location_count=self._init_kwargs.pop('min_location_count', None),
                avg_location_count=self._init_kwargs.pop('avg_location_count', None),
                max_location_count=self._init_kwargs.pop('max_location_count', None),
                )
        for k in self._explicit_kwargs:
            if k in kwargs:
                kwargs[k] = self._explicit_kwargs[k]
        helper.parse_args(kwargs)
        #
        for k,v in helper.tessellation_kwargs.items():
            #if v is None:
            #    continue
            if k in self._tessellate_kwargs:
                self._tessellate_kwargs[k] = v
            self._init_kwargs[k] = v
        if helper.scaler is not None:
            self._init_kwargs['scaler'] = helper.scaler
        #
        assert isinstance(self._tesseller, type)
        self._tesseller = (self.tesseller, helper.colnames)
    @analysis
    def tessellate(self, spt_dataframe):
        """ Grows and returns the tessellation.
        """
        if not isinstance(self._tesseller, tuple):
            self.calibrate(spt_dataframe)
        tesseller = deepcopy(self.tesseller)
        tesseller.tessellate(spt_dataframe[self.colnames], **self._tessellate_kwargs)
        #
        if self.post_processing.initialized:
            tesseller = self.post_processing.post_process(tesseller, spt_dataframe[self.colnames])
        #
        return tesseller
    @property
    def colnames(self):
        if not isinstance(self._tesseller, tuple):
            raise RuntimeError('the tesseller is not calibrated')
        return self._tesseller[1]
    def bc_update_params(self, params):
        if self.alg_name is not None:
            params['method'] = self.alg_name
        if self._init_kwargs:
            exclude = ('scaler',)
            params['tessellation'] = { k:self._init_kwargs[k]
                    for k in self._init_kwargs
                    if k not in exclude }

    scaler          = proxy_property('scaler',           'attr')
    ref_distance    = proxy_property('ref_distance', '__init__')

    @property
    def resolution(self):
        if self.ref_distance is None:
            return None
        else:
            return 2. * self.ref_distance
    @resolution.setter
    def resolution(self, res):
        if res is None:
            self.ref_distance = None
        else:
            self.ref_distance = .5 * res

    def _get_post_processing(self):
        return self._post_processing
    def _set_post_processing(self, merger):
        self._post_processing = merger
    post_processing = selfinitializing_property('post_processing', _get_post_processing, _set_post_processing, TessellationPostProcessing)

    @property
    def _mpl_impl(self):
        from .mpl import Mpl
        return Mpl
    @property
    def mpl(self):
        """ tramway.analyzer.tesseller.mpl.Mpl: Matplotlib utilities """
        return self._mpl_impl(self)

    def __getattr__(self, attrname):
        """ Beware that it ignores :attr:`_init_kwargs` symbols;
        :meth:`__init__` arguments should be made available defining a proxy property with
        the :const:`'__init__'` flag."""
        try:
            val = self._tessellate_kwargs[attrname]
        except KeyError:
            val = getattr(self.tesseller, attrname)
        return val
    def __setattr__(self, attrname, val):
        """ Beware that it ignores :attr:`_init_kwargs` symbols.

        Important note: __slots__ must be defined instead of __dict__
            so that object.__setattr__ fails with AttributeError.
        """
        # special setters for self-initializing properties
        if attrname == 'post_processing' and isinstance(self.post_processing, Initializer):
            self.post_processing.from_callable(val)
            return
        #
        try:
            AnalyzerNode.__setattr__(self, attrname, val)
        except AttributeError:
            if attrname in ('cls','tesseller','post_processing','mpl'):
                raise AttributeError(attrname+' is read-only')
            if attrname in self._tessellate_kwargs:
                self._tessellate_kwargs[attrname] = val
            else:
                try:
                    setattr(self.tesseller, attrname, val)
                except AttributeError:
                    self._init_kwargs[attrname] = val
            self._explicit_kwargs[attrname] = val

Tesseller.register(TessellerProxy)

class FreeResolutionTesseller(TessellerProxy):
    __slots__ = ()
    @property
    def resolution(self):
        if self.ref_distance is None:
            return None
        else:
            return 2. * self.ref_distance
    @resolution.setter
    def resolution(self, res):
        if res is None:
            self.ref_distance = None
        else:
            self.ref_distance = .5 * res
            try:
                self.min_probability = 0
            except AttributeError:
                pass
            try:
                self.avg_probability = 0
            except AttributeError:
                pass


__all__ = ['TessellerProxy', 'proxy_property', 'FreeResolutionTesseller' ]




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.plot.animation import *
import numpy as np


def animate_trajectories_2d(xyt, output_file=None,
        frame_rate=None, bit_rate=None, dots_per_inch=200, play=False,
        time_step=None, time_unit='s', color=None,
        line_width=1, marker_style='o', marker_size=4,
        figure=None, axes=None, bounding_box=None, axis=True,
        verbose=True, time_precision=2, **kwargs):
    """
    Animate 2D trajectories.

    Arguments:

        xyt (pandas.DataFrame): nxyt data.

        output_file (str): path to .mp4 file.

        frame_rate (float): number of frames per second.

        bit_rate (int): movie bitrate; see also :class:`~matplotlib.animation.FFMpegWriter`.

        dots_per_inch (int): dpi argument passed to :meth:`~matplotlib.animation.FFMpegWriter.saving`.

        play (bool): play the movie once generated.

        time_step (float): time step between successive frames.

        time_unit (str): time unit for time display; if None, no title is displayed.

        color (list or numpy.ndarray or callable): trajectory colour.

        line_width (float): translocation line width.

        marker_style (str): location marker style.

        marker_size (float): location marker size.

        figure (matplotlib.figure.Figure): figure handle.

        axes (matplotlib.axes.Axes): axes handle.

        bounding_box (matplotlib.transforms.Bbox): axis bounding box.

        axis (bool or str): make the axes visible.

        verbose (bool): ask for confirmation if the output file is to be overwritten
            and draw a progress bar (module :mod:`tqdm` must be available).

        time_precision (int): number of decimals for time display.

    Extra keyword arguments are passed to :class:`~matplotlib.animation.FFMpegWriter`.

    """

    if marker_style is None:
        base_style = '-'
    else:
        base_style = marker_style + '-'
    _style = dict(linewidth=line_width, markersize=marker_size)

    if color is None:
        def _color(n):
            while True:
                _c = np.random.rand(3)
                if np.any(.1 <= _c): # not too light (background is white)
                    break
            return _c
    elif callable(color):
        _color = color
    else:
        ncolors = len(color)
        def _color(n):
            return color[np.mod(n, ncolors)]
    def style(n):
        _style['color'] = _color(n)
        return _style

    if time_unit:
        title_pattern = "time = {{:.{:d}f}} {}".format(time_precision, time_unit)

    if time_step is None:
        dt = xyt['t'].diff()[xyt['n'].diff()==0].quantile(.5)
        if verbose:
            print("selected time step: {}".format(dt))
    else:
        dt = time_step
    if frame_rate is None:
        frame_rate = 1. / dt # assume dt is in seconds
    if bounding_box is None:
        import matplotlib.transforms
        xmin, ymin = xyt['x'].min(), xyt['y'].min()
        xmax, ymax = xyt['x'].max(), xyt['y'].max()
        bounding_box = matplotlib.transforms.Bbox.from_bounds(xmin, ymin, xmax-xmin, ymax-ymin)

    t0 = np.round(xyt['t'].min() / dt) * dt
    N = np.round((xyt['t'].max() - t0) / dt) + 1

    pending = xyt.copy()
    active = dict()
    old_active = set()

    try:
        with VideoWriterReader(output_file, frame_rate, bit_rate, dots_per_inch,
                figure, axes, axis, verbose, **kwargs) as movie:

            with movie.saving():
                for f in movie.range(N):
                    t = t0 + f * dt
                    xt = pending[np.abs(pending['t'] - t) < .5 * dt]
                    new_active = set()
                    for i, row in xt.iterrows():
                        n = row['n']
                        x, y = row['x'], row['y']
                        new_active.add(n)
                        try:
                            line = active[n]
                        except KeyError:
                            line, = movie.axes.plot(x, y, base_style, **style(n))
                            active[n] = line
                        else:
                            _x, _y = line.get_data()
                            line.set_data(np.r_[_x, x], np.r_[_y, y])
                    for n in old_active - new_active:
                        movie.axes.lines.remove(active.pop(n))
                    #
                    if time_unit:
                        movie.axes.set_title(title_pattern.format(t))
                    if bounding_box is not None:
                        movie.axes.dataLim.set(bounding_box)
                    #
                    movie.grab_frame()
                    #
                    old_active = new_active
                    pending.drop(xt.index, inplace=True)

            if play:
                movie.play()

    except Aborted:
        pass


__all__ = ['animate_trajectories_2d']




# -*- coding: utf-8 -*-

# Copyright © 2017-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import *
from tramway.core.exceptions import *
from tramway.tessellation import format_cell_index, nearest_cell
import tramway.tessellation as tessellation
from .gradient import grad1, delta0
import numpy as np
import pandas as pd
import scipy.sparse as sparse
import scipy.spatial.qhull
from copy import copy
from collections import OrderedDict

try:
    import cloudpickle
except ImportError:
    pass
from multiprocessing import Pool, Lock

import os # for os.name
import six
from functools import partial
from warnings import warn
import traceback



class Local(Lazy):
    """
    Spatially local subset of elements (e.g. translocations). Abstract class.

    Attributes:

        index (int):
            This cell's index as referenced in :class:`FiniteElements`.

        data (collection of terminal elements or :class:`Local`):
            Elements, either terminal or not.

        center (array-like):
            Cell center coordinates.

        span (array-like):
            Difference vectors from this cell's center to adjacent centers.

        boundary/hull (scipy.spatial.qhull.ConvexHull):
            Convex hull of the contained locations.

    """
    __slots__ = ('index', 'data', 'center', 'span', '_boundary', '_volume')
    __lazy__ = Lazy.__lazy__ + ('volume',)

    def __init__(self, index, data, center=None, span=None, boundary=None):
        Lazy.__init__(self)
        self.index = index
        self.data = data
        self.center = center
        self.span = span
        self._boundary = boundary
        self._volume = None

    @property
    def boundary(self):
        """
        *any*, property

        Cell boundary/hull.
        """
        return self._boundary

    @boundary.setter
    def boundary(self, b):
        self.volume = None
        self._boundary = b

    @property
    def hull(self):
        """
        `scipy.spatial.qhull.ConvexHull` (or *any* if not convex), property

        Alias for `boundary`.
        """
        return self.boundary

    @hull.setter
    def hull(self, h):
        self.boundary = h

    @property
    def volume(self):
        """
        `float`, property

        Surface area (2D) or volume (3D+) of the convex hull.

        Note that the hull should be convex so that the volume can be
        transparently extracted.
        In other cases, `volume` can be manually set just like a normal
        attribute.
        """
        if self._volume is None and self.hull is not None:
            if self.dim == 2:
                area = 0
                u = self.center
                for i in self.hull.simplices:
                    v, w = self.hull.points[i]
                    area += .5 * abs(\
                            (v[0] - u[0]) * (w[1] - u[1]) - \
                            (w[0] - u[0]) * (v[1] - u[1]) \
                        )
                self._volume = area
            else:
                self._volume = self.hull.volume
        return self.__returnlazy__('volume', self._volume)

    @volume.setter
    def volume(self, v):
        self.__setlazy__('volume', v)

    @property
    def dim(self):
        """
        `int`, property

        Dimension of the terminal elements.
        """
        raise NotImplementedError('abstract method')

    @dim.setter
    def dim(self, d):
        self.__assertlazy__('dim', d, related_attribute='data')

    @property
    def tcount(self):
        """
        `int`, property

        Total number of terminal elements (e.g. translocations).
        """
        raise NotImplementedError('abstract method')

    @tcount.setter
    def tcount(self, c):
        self.__assertlazy__('tcount', c, related_attribute='data')

    def get_localization_error(self, _kwargs=None, _default_value=None, _localization_error_is_sigma=False, **kwargs):
        """
        Return the localization error as :math:`sigma^2`.

        Arguments:

            _kwargs (dict):
                mutable keyword arguments;
                '*localization_error*', '*sigma*' and '*sigma2*' are popped out.

            _default_value (float):
                default localization error (subject to `_localization_error_is_sigma`).

            _localization_error_is_sigma (bool):
                argument '*localization_error*' in `_kwargs` or `kwargs` is considered as
                :math:`sigma`.

        Returns:

            float: :math:`sigma^2`.

        """
        if _localization_error_is_sigma:
            _default_sigma2 = _default_value * _default_value
        else:
            _default_sigma2 = _default_value

        if _kwargs:
            args = {}
            for arg in ('localization_error', 'sigma', 'sigma2'):
                try:
                    val = kwargs[arg]
                except KeyError:
                    try:
                        val = _kwargs.pop(arg)
                    except KeyError:
                        val = None
                args[arg] = val
        elif kwargs:
            args = kwargs
        else:
            return _default_sigma2

        localization_error = args.get('localization_error', None)
        sigma = args.get('sigma', None)
        sigma2 = args.get('sigma2', None)

        if localization_error is None:
            _sigma2 = None
        elif _localization_error_is_sigma:
            _sigma2 = localization_error * localization_error
        else:
            _sigma2 = localization_error

        if sigma2 is None:
            if sigma is None:
                if localization_error is None:
                    sigma2 = _default_sigma2
                else:
                    if _localization_error_is_sigma:
                        warn('`localization_error` may become sigma square in the coming 0.4 release, instead of sigma; please use `sigma` or `sigma2` to disambiguate', PendingDeprecationWarning)
                    sigma2 = _sigma2
            else:
                sigma2 = sigma * sigma
                if not (localization_error is None or np.isclose(_sigma2, sigma2)):
                    raise MultipleArgumentError('localization_error', 'sigma')
        else:
            if not (sigma is None or np.isclose(sigma*sigma, sigma2)):
                if localization_error is None:
                    raise MultipleArgumentError('sigma', 'sigma2')
                else:
                    raise MultipleArgumentError('localization_error', 'sigma', 'sigma2')
            if not (_sigma2 is None or np.isclose(_sigma2, sigma2)):
                raise MultipleArgumentError('localization_error', 'sigma2')

        return sigma2



class Distributed(Local):
    """
    Attributes:

        central (array of bools):
            margin cells are not central.

    """
    __slots__ = ('_reverse', '_adjacency', 'central', '_degree', '_ccount', '_tcount')
    __lazy__  = Local.__lazy__ + ('reverse', 'degree', 'ccount', 'tcount')

    def __init__(self, cells, adjacency, index=None, center=None, span=None, central=None, \
        boundary=None):
        Local.__init__(self, index, OrderedDict(), center, span, boundary)
        self.cells = cells # let's `cells` setter perform the necessary checks
        self.adjacency = adjacency
        self.central = central

    @property
    def cells(self):
        """
        `list` or `OrderedDict`, rw property for :attr:`data`

        Collection of :class:`Local`. Indices may not match with the global
        :attr:`~Local.index` attribute of the elements, but match with attributes
        :attr:`central`, :attr:`adjacency` and :attr:`degree`.
        """
        return self.data

    @cells.setter
    def cells(self, cells):
        celltype = type(self.cells)
        assert(celltype is dict or celltype is OrderedDict)
        if not isinstance(cells, celltype):
            if isinstance(cells, dict):
                cells = celltype(sorted(cells.items(), key=lambda t: t[0]))
            elif isinstance(cells, list):
                cells = celltype(sorted(enumerate(cells), key=lambda t: t[0]))
            else:
                raise TypeError('`cells` argument is not a dictionnary (`dict` or `OrderedDict`)')
        if not all([ isinstance(cell, Local) for cell in cells.values() ]):
            raise TypeError('`cells` argument is not a dictionnary of `Local`')
        #try:
        #       if self.ccount == len(cells): #.keys().reversed().next(): # max
        #           self.adjacency = None
        #           self.central = None
        #except:
        #       pass
        self.reverse = None
        self.ccount = None
        self.data = cells

    @property
    def indices(self):
        return np.array([ cell.index for cell in self.cells.values() ])

    @property
    def reverse(self):
        """
        `dict of ints`, ro lazy property

        Get "local" indices from global ones.
        """
        if self._reverse is None:
            self._reverse = {cell.index: i for i, cell in self.cells.items()}
        return self._reverse

    @reverse.setter
    def reverse(self, r): # ro
        self.__assertlazy__('reverse', r, related_attribute='cells')

    @property
    def space_cols(self):
        """
        `list of ints` or `strings`, ro property

        Column indices for coordinates of the terminal points.
        """
        return self.any_cell().space_cols

    @property
    def dim(self):
        """
        `int`, ro property

        Dimension of the terminal points.
        """
        if self.center is None:
            return self.any_cell().dim
        else:
            return self.center.size

    @property
    def tcount(self):
        """
        `int`, ro property

        Total number of terminal points. Duplicates are ignored.
        """
        if self._tcount is None:
            if self.central is None:
                self._tcount = sum([ cell.tcount \
                    for i, cell in self.cells.items() if self.central[i] ])
            else:
                self._tcount = sum([ cell.tcount for cell in self.cells.values() ])
        return self._tcount

    @tcount.setter
    def tcount(self, c):
        # write access allowed for performance issues, but `c` should equal self.tcount
        self.__setlazy__('tcount', c)

    @property
    def ccount(self):
        """
        `int`, ro property

        Total number of terminal cells. Duplicates are ignored.
        """
        #return self.adjacency.shape[0] # or len(self.cells)
        if self._ccount is None:
            self._ccount = sum([ cell.ccount if isinstance(cell, FiniteElements) else 1 \
                for cell in self.cells.values() ])
        return self._ccount # not mutable (no need for __returnlazy__)

    @ccount.setter
    def ccount(self, c): # rw for performance issues, but `c` should equal self.ccount
        self.__setlazy__('ccount', c)

    @property
    def adjacency(self):
        """
        `scipy.sparse.csr_matrix`, rw property

        Cell adjacency matrix. Row and column indices are to be mapped with `indices`.
        """
        return self._adjacency

    @adjacency.setter
    def adjacency(self, a):
        if a is not None:
            a = a.tocsr()
        self._adjacency = a
        self._degree = None # `degree` is ro, hence set `_degree` instead

    @property
    def degree(self):
        """
        `array of ints`, ro lazy property

        Number of adjacent cells.
        """
        if self._degree is None:
            self._degree = np.diff(self._adjacency.indptr)
        return self._degree

    @degree.setter
    def degree(self, d): # ro
        self.__lazyassert__(d, 'adjacency')

    def grad(self, i, X, index_map=None, **kwargs):
        """
        Local spatial gradient.

        Sub-classes are free to return multi-component gradients as matrices provided that
        they exhibit as many columns as there are dimensions in the (trans-)location data.
        :meth:`grad_sum` is then responsible for summing all the elements of such matrices.

        Arguments:

            i (int):
                cell index at which the gradient is evaluated.

            X (numpy.ndarray):
                vector of a scalar measurement at every cell.

            index_map (numpy.ndarray):
                index map that converts cell indices to indices in X.

        Returns:

            numpy.ndarray:
                gradient vector with as many elements as spatial dimensions.

        See also :func:`~tramway.inference.gradient.grad1` and documentation section :ref:`gradient`.
        """
        return grad1(self, i, X, index_map, **kwargs)

    def grad_sum(self, i, grad2, index_map=None):
        """
        Sum for local penalties (*e.g.* local gradient vector with squared
        elements) at a given cell.

        Arguments:

            i (int):
                cell index.

            grad2 (numpy.ndarray):
                local penalties (*e.g.* local gradient vector with squared
                elements).

            index_map (numpy.ndarray):
                index mapping, useful to convert cell indices to positional indices in
                an optimization array for example.

        Returns:

            float:
                weighted sum of the elements of `grad2`.

        """
        cell = self.cells[i]
        if cell.volume:
            return cell.volume * np.sum(grad2)
        else:
            return np.sum(grad2)

    def local_variation(self, i, X, index_map=None, **kwargs):
        """
        Local spatial variation, gradient-like, aimed at penalizing spatial variations.

        See also :func:`~tramway.inference.gradient.delta0`.

        As of version *0.3.8*: new; called for spatial regularization in `stochastic_dv`.

        As of version *0.4*: default implementation becomes :func:`~tramway.inference.gradient.delta0`.

        May become the new default for spatial regularization.

        Arguments:

            i (int):
                cell index at which the gradient is evaluated.

            X (numpy.ndarray):
                vector of a scalar measurement at every cell.

            index_map (numpy.ndarray):
                index map that converts cell indices to indices in X.

        Returns:

            numpy.ndarray:
                delta vector with as many elements as there are spatial dimensions.

        """
        return delta0(self, i, X, index_map, **kwargs)

    def flatten(self):
        def concat(arrays):
            if isinstance(arrays[0], tuple):
                raise NotImplementedError
            elif isinstance(arrays[0], pd.DataFrame):
                return pd.concat(arrays, axis=0)
            else:
                return np.stack(arrays, axis=0)

        new = copy(self)
        new.cells = {i: Cell(i, concat([cell.data for cell in dist.cells.values()]), \
                dist.center, dist.span, hull=dist.hull) \
            if isinstance(dist, FiniteElements) else dist \
            for i, dist in self.cells.items() }

        return new

    def group(self, ngroups=None, max_cell_count=None, cell_centers=None, \
        adjacency_margin=2, connected=False):
        """
        Make groups of cells.

        This builds up an extra hierarchical level.
        For example if `self` is a `FiniteElements` of `Cell`, then this returns a `FiniteElements`
        of `FiniteElements` (the groups) of `Cell`.

        Several grouping strategies are proposed.

        Arguments:

            ngroups (int):
                number of groups.

            max_cell_count (int):
                maximum number of cells per group.

            cell_centers (array-like):
                spatial centers of the groups.
                Cells are sorted by nearest neighbours.
                If not provided, :meth:`group` will use
                a k-means approach to positionning the centers.

            adjacency_margin (int):
                groups are dilated to the adjacent cells `adjacency_margin` times.
                Defaults to 2.

            connected (bool):
                separates the connected components from one another.
                Conflicts with the other arguments.

        Returns:

            FiniteElements:
                a new object of the same type as `self` that contains other such
                objects as cells.

        """
        ncells = self.adjacency.shape[0]
        new = copy(self)

        strategy_0 = ngroups or (max_cell_count and max_cell_count < ncells) or cell_centers is not None
        if not strategy_0 and not connected:
            #raise KeyError('`group` expects more input arguments')
            return new

        elif strategy_0:

            if connected:
                raise ValueError('`connected` is not supported in combination with other arguments')

            if cell_centers is None and max_cell_count == 1 and not adjacency_margin:
                distr = type(new)
                for i in new:
                    cell = new[i]
                    new[i] = distr({i: cell}, sparse.eye(1, dtype=bool, format='csr'),
                            center=cell.center)
                return new

            any_cell = self.any_cell()

            points = np.zeros((ncells, self.dim), dtype=any_cell.center.dtype)
            ok = np.zeros(points.shape[0], dtype=bool)
            for i in self.cells:
                if self.cells[i]: # non-empty
                    points[i] = self.cells[i].center
                    ok[i] = True

            if cell_centers is None:
                if max_cell_count == 1:
                    assert False
                    grid = tessellation.Voronoi()
                    grid.cell_centers = points
                else:
                    avg_probability = 1.0
                    if ngroups:
                        avg_probability = min(1.0 / float(ngroups), avg_probability)
                    if max_cell_count:
                        avg_probability = min(float(max_cell_count) / \
                            float(points.shape[0]), avg_probability)
                    import tramway.tessellation.kmeans as kmeans
                    grid = kmeans.KMeansMesh(avg_probability=avg_probability)
                    try:
                        grid.tessellate(points[ok])
                    except scipy.spatial.qhull.QhullError:
                        print('Qhull errors are handled; full map optimization')
                        return new
                    except ValueError:
                        print(points)
                        raise

            else:
                grid = tessellation.Voronoi()
                grid.cell_centers = cell_centers

            I = np.full(ok.size, -1, dtype=int)
            I[ok] = grid.cell_index(points[ok], min_location_count=1)
            #if not np.all(ok):
            #       print(ok.nonzero()[0])
            new.adjacency = grid.simplified_adjacency(format='csr') # macro-cell adjacency matrix
            J = np.unique(I)
            J = J[0 <= J]
            assert 0 < J.size
            new.data = type(self.cells)()

            for j in J: # for each macro-cell
                K = I == j # find corresponding cells
                assert np.any(K)

                if 0 < adjacency_margin:
                    L = np.copy(K)
                    for k in range(adjacency_margin):
                        # add adjacent cells for future gradient calculations
                        K[self.adjacency[K,:].indices] = True
                    L = L[K]

                A = self.adjacency[K,:].tocsc()[:,K].tocsr() # point adjacency matrix
                C = grid.cell_centers[j]
                D = OrderedDict([ (i, self.cells[k]) \
                    for i, k in enumerate(K.nonzero()[0]) if k in self.cells ])

                for i in D:
                    adj = A[i].indices
                    if 0 < D[i].tcount and adj.size:
                        span = np.stack([ D[k].center for k in adj ], axis=0)
                    else:
                        span = np.empty((0, D[i].center.size), \
                            dtype=D[i].center.dtype)
                    if span.shape[0] < D[i].span.shape[0]:
                        D[i] = copy(D[i])
                        D[i].span = span - D[i].center

                R = grid.cell_centers[new.adjacency[j].indices] - C

                new.cells[j] = type(self)(D, A, index=j, center=C, span=R)

                if 0 < adjacency_margin:
                    new.cells[j].central = L
                assert bool(new.cells[j])

        else:
            new.cells = type(self.cells)()
            j = 0
            available_cells = self.cells
            while available_cells:

                any_cell = next(iter(available_cells))
                neighbours = set(self.neighbours(any_cell).tolist())
                visited_cells = set([any_cell])
                while neighbours:
                    visited_cells |= neighbours
                    new_neighbours = set()
                    for cell in neighbours:
                        new_neighbours |= set(self.neighbours(cell).tolist())
                    neighbours = new_neighbours - visited_cells
                available_cells = { k: cell for k, cell in available_cells.items()
                    if k not in visited_cells } # update for next iteration

                K = list(visited_cells)
                A = self.adjacency[K,:].tocsc()[:,K].tocsr() # point adjacency matrix
                D = OrderedDict([ (i, self.cells[k]) \
                    for i, k in enumerate(K) if k in self.cells ])

                new.cells[j] = type(self)(D, A, index=j)

                j += 1

            new.adjacency = sparse.csr_matrix((np.array([], dtype=bool), (np.array([], dtype=int), np.array([], dtype=int))), shape=(j, j))


        new.ccount = self.ccount
        # _tcount is not supposed to change

        return new

    def run(self, function, *args, **kwargs):
        """
        Apply a function that takes a group (:class:`FiniteElements`) of terminal cells
        as input argument, plus args and kwargs, and must return a `pandas.DataFrame` array
        with the indices referring to cells.

        `function` is called for each group of terminal cells, adjacency margins
        are removed if any, and the resulting `DataFrame` are merged into a single
        `DataFrame`.

        `function` may instead be applied to `self` in the case `self.cells` has been
        overloaded to exhibit the output features as attributes.

        Arguments:

            function (callable):
                the function to be called on each terminal :class:`FiniteElements`.
                Its first argument is the :class:`FiniteElements` object.
                It should return maps as a :class:`~pandas.DataFrame` and optionally
                posteriors as :class:`~pandas.Series` or :class:`~pandas.DataFrame`.

            args (list):
                positional arguments for `function` after the first one.

            kwargs (dict):
                keyword arguments for `function`;
                the following arguments are popped out of `kwargs`:

            returns (list):
                attributes to be collected from all the individual cells as return values;
                if defined, the values returned by `function` are ignored.

            worker_count (int):
                number of simultaneously working processing units,
                if the `self.cells` .

            profile (bool or str or tuple):
                profile each child job if any;
                if `str`, dump the output stats into *.prof* files;
                if `tuple`, print a report with :func:`~pstats.Stats.print_stats` and
                tuple elements as input arguments.

            function_worker_count (int):
                if `worker_count` is a reserved keyword argument for `function`,
                `function_worker_count` can be specified and is passed to `function`
                with name/keyword `worker_count`;
                in this case, `worker_count` must be a multiple of `function_worker_count`.

        Returns:

            pandas.DataFrame:
                single merged array of maps.
                If `function` returns two output arguments, :meth:`run` also
                returns a second merged array of posterior probabilities(?).

        """
        # clear the caches
        self.clear_caches()

        returns = kwargs.pop('returns', None)

        parallel = all(isinstance(cell, FiniteElements) for cell in self.cells.values())
        if parallel:
            worker_count = kwargs.pop('worker_count', None)
            profile = kwargs.pop('profile', False)

        for arg in ('returns', 'worker_count', 'profile'):
            try:
                val = kwargs.pop('function_'+arg)
            except KeyError:
                pass
            else:
                kwargs[arg] = val
                if arg == 'worker_count' and \
                        isinstance(worker_count, int) and \
                        isinstance(val, int) and 0<val:
                    worker_count /= val

        if parallel:
            # parallel for-loop over the subsets of cells

            ## if Windows, make the computation sequential
            #if os.name == 'nt':
            #    if worker_count is None:
            #        worker_count = 0
            #    else:
            #        warn('multiprocessing may break on Windows', RuntimeWarning)

            # if `worker_count` is `None`, `Pool` will use `multiprocessing.cpu_count()`
            # if `worker_count == 0`, make it single-processing
            if worker_count == 0:
                pool = None
            else:
                pool = Pool(worker_count)
            fargs = (function, args, kwargs)
            if profile:
                fargs = (profile, fargs)
                cells = [ (i, self.cells[i]) for i in self.cells ]#if bool(self.cells[i]) ]
            else:
                cells = [ self.cells[i] for i in self.cells ]#if bool(self.cells[i]) ]
            if six.PY3 or pool is None:
                if profile:
                    _run = __profile_run__
                else:
                    _run = __run__
                if pool is None:
                    ys = [ _run(fargs, c) for c in cells ]
                else:
                    ys = pool.map(partial(_run, fargs), cells)
            elif six.PY2:
                import itertools
                if profile:
                    _run = __profile_run_star__
                else:
                    _run = __run_star__
                ys = pool.map(_run,
                    itertools.izip(itertools.repeat(fargs), cells))
            if returns is None:
                ys = [ y for y in ys if y is not None ]
                if ys:
                    if ys[1:]:
                        if isinstance(ys[0], tuple):
                            ys = zip(*ys)
                            result = tuple([
                                pd.concat(_ys, axis=0).sort_index()
                                for _ys in ys
                                if _ys and isinstance(_ys[0], pd.DataFrame) ])
                            if not result[1:]:
                                result, = result
                        else:
                            result = pd.concat(ys, axis=0).sort_index()
                    else:
                        result = ys[0]
                else:
                    result = None

        else:
            # direct function application
            result = function(self, *args, **kwargs)

        if returns:
            index = { v: [] for v in returns }
            result = { v: [] for v in returns }
            for i in self.cells:
                cell = self.cells[i]
                if cell:
                    for v in returns:
                        try:
                            x = getattr(cell, v)
                        except AttributeError:
                            x = None
                        if x is not None:
                            index[v].append(i)
                            result[v].append(x)
            for v in result:
                x = np.vstack(result[v])
                result[v] = pd.DataFrame(x, index=index[v], \
                        columns=[v] if x.shape[1] == 1 \
                        else [ '{} {:d}'.format(v, i) for i in range(x.shape[1]) ])
            _result = result
            result = _result[returns[0]]
            if returns[1:]:
                result = result.join([ _result[v] for v in returns[1:] ])

        return result

    # `dict` interface
    def __len__(self):
        return self.adjacency.shape[0]

    def __nonzero__(self):
        return bool(self.cells)

    def __iter__(self):
        return iter(self.cells)

    def __getitem__(self, i):
        return self.cells[i]

    def __setitem__(self, i, cell):
        self.cells[i] = cell

    def __delitem__(self, i):
        self.cells.__delitem__(i)

    def keys(self):
        try:
            return self.cells.keys()
        except AttributeError:
            return range(len(self.cells))

    def values(self):
        try:
            return self.cells.values()
        except AttributeError:
            return self.cells

    def items(self):
        try:
            return self.cells.items()
        except AttributeError:
            return enumerate(self.cells)

    def any_cell(self):
        return self.cells[next(iter(self.cells))]

    def neighbours(self, i):
        """
        Indices of neighbour cells.

        Argument:

            i (int): cell index.

        Returns:

            numpy.ndarray: indices of the neighbour cells of cell *i*.

        """
        return self.adjacency.indices[self.adjacency.indptr[i]:self.adjacency.indptr[i+1]]

    def clear_caches(self):
        try:
            first = True
            for c in self.values():
                c.clear_cache()
                first = False
        except AttributeError as e:
            if first:
                try:
                    first = True
                    for c in self.values():
                        c.clear_caches()
                        first = False
                    return
                except:
                    if first:
                        raise e
            if not first:
                raise


def __run__(func, cell):
    function, args, kwargs = func
    try:
        x = cell.run(function, *args, **kwargs)
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        print(traceback.format_exc())
        raise
    if x is None:
        return None
    else:
        is_tuple = isinstance(x, tuple)
        if is_tuple:
            _x = x[1:]
            x = x[0]
        i = cell.indices
        if cell.central is not None:
            try:
                x = x.iloc[cell.central[x.index]]
            except IndexError as e:
                if cell.central.size < x.index.max():
                    raise IndexError('dataframe indices do no match with group-relative cell indices (maybe are they global ones)')
                else:
                    print(x.shape)
                    print((cell.central.shape, cell.central.max()))
                    print(x.index.max())
                    raise e
            i = i[x.index]
            if x.shape[0] != i.shape[0]:
                raise IndexError('not as many indices as values')
        x.index = i
        if is_tuple:
            return (x,)+_x
        else:
            return x

def __run_star__(args):
    return __run__(*args)


def __profile_run__(func, args):
    import cProfile, pstats
    proptions, func = func
    try:
        process, cells = args
    except TypeError:
        process = None
    profile = cProfile.Profile()
    profile.enable()
    result = __run__(func, cells)
    profile.disable()
    if process is not None and isinstance(proptions, str):
        profile.dump_stats('{}{}.prof'.format(filename, process))
    else:
        if proptions in (True, False):
            proptions = (.1, )
        elif not isinstance(proptions, (tuple, list)):
            proptions = (proptions, )
        stats = pstats.Stats(profile).sort_stats('cumulative')
        stats.print_stats(*proptions)
    return result

def __profile_run_star__(args):
    return __profile_run__(*args)


FiniteElements = Distributed



class Cell(Local):
    """
    Spatially constrained subset of (trans-)locations with associated intermediate calculations.

    Attributes:

        cache (any):
            Depending on the inference approach and objective, caching an intermediate
            result may avoid repeating many times a same computation. Usage of this cache
            is totally free and comes without support for concurrency.

    """
    __slots__ = ('_time_col', '_space_cols', 'cache', 'fuzzy')
    __lazy__  = Local.__lazy__ + ('time_col', 'space_cols')

    def __init__(self, index, data, center=None, span=None, boundary=None):
        """
        Arguments:

            index (int):
                this cell's index as granted in :class:`FiniteElements` 's `cells` dict.

            center (array-like):
                cell center coordinates.

            span (array-like):
                difference vectors from this cell's center to adjacent centers.

        """
        if not (isinstance(data, np.ndarray) or isinstance(data, pd.DataFrame)):
            raise TypeError('unsupported (trans-)location type `{}`'.format(type(data)))
        Local.__init__(self, index, data, center, span, boundary)
        #self._tcount = data.shape[0]
        self._time_col = None
        self._space_cols = None
        self.cache = None
        #self.data = (self.space_data, self.time_data)
        self.fuzzy = None

    @property
    def time_col(self):
        """
        `int` or `string`, lazy

        Column index for time.
        """
        if self._time_col is None:
            if isstructured(self.data):
                self._time_col = 't'
            else:
                self._time_col = 0
        return self._time_col

    @time_col.setter
    def time_col(self, col):
        if isinstance(col, (tuple, list)):
            if not col or col[1:]:
                raise ValueError('a single time column is supported')
            col = col[0]
        # space_cols is left unchanged
        self.__lazysetter__(col)

    @property
    def space_cols(self):
        """
        `list of ints` or `strings`, lazy

        Column indices for coordinates.
        """
        if self._space_cols is None:
            if isstructured(self.data):
                self._space_cols = columns(self.data)
                if isinstance(self._space_cols, pd.Index):
                    self._space_cols = self._space_cols.drop(self.time_col)
                else:
                    self._space_cols.remove(self.time_col)
            else:
                if self.time_col == 0:
                    self._space_cols = np.arange(1, self.data.shape[1])
                else:
                    self._space_cols = np.ones(self.data.shape[1], \
                        dtype=bool)
                    self._space_cols[self.time_col] = False
                    self._space_cols, = self._space_cols.nonzero()
        return self._space_cols

    @space_cols.setter
    def space_cols(self, cols):
        # time_col is left unchanged
        self.__lazysetter__(cols)

    @property
    def dim(self):
        return len(self.space_cols)

    @dim.setter
    def dim(self, d): # ro
        self.__lazyassert__(d, 'data')

    @property
    def tcount(self):
        """
        `int`

        Number of (trans-)locations.
        """
        return self.space_data.shape[0]

    @tcount.setter
    def tcount(self, c): # ro
        self.__lazyassert__(c, 'data')

    @property
    def time_data(self):
        if not isinstance(self.data, tuple):
            self.data = (self._extract_space(), self._extract_time())
        return self.data[1]

    @time_data.setter
    def time_data(self, t):
        self.data = (self.space_data, t)

    def _extract_time(self):
        if isstructured(self.data):
            return np.asarray(self.data[self.time_col])
        else:
            return np.asarray(self.data[:,self.time_col])

    @property
    def space_data(self):
        if not isinstance(self.data, tuple):
            self.data = (self._extract_space(), self._extract_time())
        return self.data[0]

    @space_data.setter
    def space_data(self, xy):
        self.data = (xy, self.time_data)

    def _extract_space(self):
        if isstructured(self.data):
            return np.asarray(self.data[self.space_cols])
        else:
            return np.asarray(self.data[:,self.space_cols])

    def __len__(self):
        return self.tcount

    def __nonzero__(self):
        if isinstance(self.data, tuple):
            return 0 < int(self.data[0].size)
        else:
            return 0 < int(self.data.size)

    def clear_cache(self):
        self.cache = None


FiniteElement = Cell



class Locations(Cell):
    """

    """
    __slots__ = ()

    @property
    def locations(self):
        """
        `array-like`, property

        Locations as a matrix of coordinates and times with as many
        columns as dimensions; this is an alias for :attr:`~Local.data`.
        """
        return self.data

    @locations.setter
    def locations(self, tr):
        self.data = tr

    @property
    def t(self):
        """
        `array-like`, ro property

        Location timestamps.
        """
        return self.time_data

    @t.setter
    def t(self, t):
        self.time_data = t

    @property
    def r(self):
        """
        `array-like`, ro property

        Location coordinates; `xy` is an alias
        """
        return self.space_data

    @r.setter
    def r(self, r):
        self.space_data = r

    @property
    def xy(self):
        return self.space_data

    @xy.setter
    def xy(self, xy):
        self.space_data = xy


class Translocations(Cell):
    """
    Attributes:

        origins (array-like, ro property):
            Initial locations (both spatial coordinates and times).

        destinations (array-like, ro property):
            Final locations (both spatial coordinates and times).

    """
    __slots__ = ('origins', 'destinations')

    def __init__(self, index, translocations, center=None, span=None, boundary=None):
        Cell.__init__(self, index, translocations, center, span, boundary)
        self.origins = None
        self.destinations = None

    @property
    def translocations(self):
        """
        `array-like`, property

        Translocations as a matrix of variations of coordinate and time with as many
        columns as dimensions; this is an alias for :attr:`~Local.data`.
        """
        return self.data

    @translocations.setter
    def translocations(self, tr):
        self.data = tr

    @property
    def dt(self):
        """
        `array-like`, ro property

        Translocation durations.
        """
        return self.time_data

    @dt.setter
    def dt(self, dt):
        self.time_data = dt

    @property
    def dr(self):
        """
        `array-like`, ro property

        Translocation displacements in space; `dxy` is an alias.
        """
        return self.space_data

    @dr.setter
    def dr(self, dr):
        self.space_data = dr

    @property
    def dxy(self):
        return self.space_data

    @dxy.setter
    def dxy(self, dxy):
        self.space_data = dxy

    @property
    def t(self):
        """
        `array-like`, ro property

        Initial timestamps.
        """
        if isstructured(self.origins):
            return np.asarray(self.origins[self.time_col])
        else:
            return np.asarray(self.origins[:,self.time_col])

    @property
    def r(self):
        if isstructured(self.origins):
            return np.asarray(self.origins[self.space_cols])
        else:
            return np.asarray(self.origins[:,self.space_cols])


class TrackedMolecules(Translocations):
    """
    Attributes:

        n (numpy.ndarray):
            Trajectory indices.

    """
    __slots__ = 'n',


def identify_columns(points, trajectory_col=True):
    """
    Identify columns by type.

    Arguments:
        points (array-like):
            location coordinates and trajectory index.

        trajectory_col (bool or int or str):
            trajectory column index or name,
            or ``False`` or ``None`` if there is no trajectory information is to be
            found or extracted,
            or ``True`` to automatically identify such a column.

    Returns:
        tuple: (*coord_cols*, *trajectory_col*, *get_var*, *get_point*)

    *coord_cols* is an array of column indices/names other than the trajectory column.
    If the data contains typical translocation column names such as 'dx', 'dy', 'dt',
    *coord_cols* is a tuple with, as first element, an array of location column names,
    and as second element an array of translocation column names.

    *trajectory_col* is the column index/name that contains tracking information.

    *get_var* is a callable that takes an array like `points` and column identifiers and
    returns the corresponding sub-matrix.

    *get_point* is a callable that takes an array like `points` and row indices and
    returns the corresponding sub-matrix.

    """
    _has_trajectory = trajectory_col not in [False, None]
    _traj_undefined = trajectory_col is True
    _has_delta_columns = False

    if isstructured(points):
        coord_cols = columns(points)
        if _has_trajectory:
            if _traj_undefined:
                trajectory_col = 'n'
            try:
                if isinstance(coord_cols, pd.Index):
                    coord_cols = coord_cols.drop(trajectory_col)
                else:
                    coord_cols.remove(trajectory_col)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                if _traj_undefined:
                    _has_trajectory = False
                else:
                    raise
        # assume no column name is empty
        delta_cols = [ col for col in coord_cols if col[0] == 'd' and col[1:] in coord_cols ]
        if delta_cols:
            coord_cols = [ col for col in coord_cols if col not in delta_cols ]
            coord_cols = (coord_cols, delta_cols)
    else:
        if isinstance(points, (tuple, list)):
            points = np.asarray(points)
        if not _has_trajectory:
            coord_cols = np.arange(points.shape[1])
        elif _traj_undefined:
            trajectory_col = 0
            coord_cols = np.arange(1, points.shape[1])
        else:
            coord_cols = np.r_[np.arange(trajectory_col), \
                np.arange(trajectory_col+1, points.shape[1])]
    if not _has_trajectory:
        trajectory_col = None

    if isinstance(points, pd.DataFrame):
        def get_point(a, i):
            return a[i]
        def get_var(a, j):
            return a[j]
    elif isstructured(points):
        def get_point(a, i):
            return a[i,:]
        def get_var(a, j):
            return a[j]
    else:
        def get_point(a, i):
            return a[i]
        def get_var(a, j):
            return a[:,j]

    return coord_cols, trajectory_col, get_var, get_point


def get_locations(points, index=None, coord_cols=None, get_var=None, get_point=None):
    """
    Make helpers for manipulating the point data.

    Arguments:

        points (array-like):
            location coordinates and trajectory index.

        index (ndarray or pair of ndarrrays or sparse matrix):
            point-cell association (see :class:`~tramway.tessellation.base.CellStats`
            documentation or :meth:`~tramway.tessellation.base.Tessellation.cell_index`).

    Returns:
        tuple: (*locations*, *location_cell*, *get_point*)

    *locations* are point coordinates (same format as `points` with the trajectory column
    removed.

    *location_cell* is either an array of cell indices (same size as *locations*) or
    a callable that takes a cell index and returns a boolean array with Trues
    for locations associated with the specified cell and Falses elsewhere.

    *get_point* is a callable that takes an array like `points` or `locations` and row indices
    and returns the corresponding rows in the same format.

    """
    if coord_cols is None:
        coord_cols, _, get_var, get_point = identify_columns(points)
    locations = get_var(points, coord_cols)
    #location_cell = format_cell_index(index, 'array', nearest_cell, (locations.shape[0],))
    if index is None or isinstance(index, np.ndarray):

        location_cell = index

    elif isinstance(index, tuple):

        _point, _cell = index
        loc_count = locations.shape[0]

        def __associated__(cell):
            """
            Location-cell association.

            Arguments:

                cell (int):
                    cell index.

            Returns:

                bool or ndarray:
                    True for locations associated to cell `cell`,
                    False otherwise.
            """
            _in = np.zeros(loc_count, dtype=bool)
            _in[_point[_cell == cell]] = True
            return _in

        location_cell = __associated__

    else:#if sparse.issparse(index):
        assert sparse.issparse(index)

        try:
            index = index.tocsc(copy=True)
        except TypeError: # not already a CSC matrix; copy is implicit
            index = index.tocsc()

        def __association__(cell):
            """
            Location-cell association.

            Arguments:

                cell (int):
                    cell index.

            Returns:

                any:
                    scalar value or ndarray which elements evaluate to
                    True for locations associated to cell `cell`,
                    False otherwise.
            """
            return index.getcol(cell).todense()

        location_cell = __association__

    return locations, location_cell, get_point


def get_translocations(points, index=None, coord_cols=None, trajectory_col=True,
        get_var=None, get_point=None):
    """
    Identify translocations as initial and final points.

    Arguments:
        points (array-like):
            location coordinates and trajectory index.

        index (ndarray or pair of ndarrays or sparse matrix):
            point-cell association (see :class:`~tramway.tessellation.base.CellStats`
            documentation or :meth:`~tramway.tessellation.base.Tessellation.cell_index`).

    Returns:
        tuple: (*initial_point*, *final_point*, *initial_cell*, *final_cell*, *get_point*)

    *initial_point* and *final_point* are point coordinates (same format as `points` with the
    trajectory index column removed) respectively of the initial and final displacement locations.

    *initial_cell* and *final_cell* are either arrays of cell indices (same size as *initial_point*)
    or callables that take a cell index and return a boolean array with Trues
    for displacements/translocations associated with the specified cell and Falses elsewhere.

    *get_point* is a callable that takes an array like `points` and row indices and returns
    the corresponding rows in the same format.

    """
    if coord_cols is None:
        coord_cols, trajectory_col, get_var, get_point = identify_columns(points, trajectory_col)
    if isinstance(coord_cols, tuple):
        coord_cols, delta_cols = coord_cols
    else:
        delta_cols = []

    if delta_cols:
        # translocation "coordinates"
        deltas = get_var(points, delta_cols)
        # location coordinates
        points = get_var(points, coord_cols)
        # fake `final`
        initial = final = ~np.any(np.isnan(np.asarray(deltas)), axis=1)#np.ones(points.shape[0], dtype=bool)
        # note: the destination cell will be undefined
    else:
        if trajectory_col is None:
            raise ValueError('cannot find trajectory indices')

        # trajectory index
        n = np.asarray(get_var(points, trajectory_col))
        # point coordinates
        points = get_var(points, coord_cols)

        initial = np.diff(n, axis=0) == 0
        final = np.r_[False, initial]
        initial = np.r_[initial, False]

    if index is None:

        initial_cell = final_cell = None

    elif isinstance(index, np.ndarray):

        initial_cell = index[initial]
        final_cell = index[final]

    elif isinstance(index, tuple):

        _point, _cell = index
        if _point.size == 0:
            raise ValueError('no data points found')
        _unique = np.unique(_point)
        loc_count = max(initial.size, _unique[-1]+1) # should be enough
        transloc_count = np.sum(initial)

        def __f__(termination):
            _loc_ok = np.zeros(loc_count, dtype=bool)
            _loc_ok[_unique] = True
            _transloc_ok = _loc_ok[termination]
            _transloc = np.arange(transloc_count)
            _transloc[~_transloc_ok] = -1
            _loc = np.full(loc_count, -1, dtype=int)
            _loc[termination] = _transloc
            _loc = _loc[_point]
            if not np.any(_transloc_ok):
                raise ValueError('no translocations available')
            def __associated__(cell):
                """
                Translocation-cell association.

                Arguments:

                    cell (int):
                        cell index.

                Returns:

                    bool or numpy.ndarray:
                        True for translocations associated to cell `cell`,
                        False otherwise.
                """
                _in = np.zeros(transloc_count, dtype=bool)
                _ok = _loc[_cell == cell]
                _in[_ok[0<=_ok]] = True
                return _in
            return __associated__

        initial_cell = __f__(initial)
        final_cell = __f__(final)
        # debug
        #_c = 0
        #pts = points[['x', 'y']].values
        #_pts = pts[_point[_cell==_c]]
        #print((_pts.min(axis=0), _pts.max(axis=0)))
        #_pts = pts[initial][initial_cell(_c)]
        #print((_pts.min(axis=0), _pts.max(axis=0)))

    else:#if sparse.issparse(index):
        assert sparse.issparse(index)

        try:
            index = index.tocsc(copy=True)
        except TypeError: # not already a CSC matrix; copy is implicit
            index = index.tocsc()
        loc_count = index.shape[0]
        transloc_count = np.sum(initial)

        def __f__(termination):
            _transloc = np.full(loc_count, -1, dtype=int)
            _transloc[termination] = np.arange(transloc_count)
            if np.all(_transloc==-1):
                raise ValueError('no translocations available')
            def __association__(cell):
                """
                Translocation-cell association.

                Arguments:

                    cell (int):
                        cell index.

                Returns:

                    any:
                        scalar value or ndarray which elements evaluate to
                        True for translocations associated to cell `cell`,
                        False otherwise.
                """
                _in = np.zeros(transloc_count, dtype=bool)
                _ok = _transloc[index.indices[index.indptr[cell]:index.indptr[cell+1]]]
                _in[_ok[0<=_ok]] = True
                return _in
            return __association__

        initial_cell = __f__(initial)
        final_cell = __f__(final)

    initial_point = get_point(points, initial)
    assert not np.any(np.isnan(np.asarray(initial_point)))

    if delta_cols:
        # if deltas are available, the destination cell is undefined
        if callable(final_cell):
            final_cell = lambda cell: []
        elif final_cell is not None:
            final_cell[...] = -1

        deltas = get_point(deltas, initial)
        cols = deltas.columns
        deltas.columns = [ col[1:].lstrip() for col in cols ]
        final_point = initial_point + deltas
        if np.any(np.isnan(np.asarray(final_point))):
            for col in initial_point.columns:
                if col not in deltas.columns:
                    if col[0] == 'd':
                        raise ValueError("no column corresponding to delta '{}'".format(col))
                    else:
                        raise ValueError("no delta column corresponding to '{}'".format(col))
            assert False # final_point contains NaNs
    else:
        final_point = get_point(points, final)
        assert not np.any(np.isnan(np.asarray(final_point)))
    assert initial_point.shape == final_point.shape

    return initial_point, final_point, initial_cell, final_cell, get_point


def distributed(cells, new_cell=None, new_group=FiniteElements, fuzzy=None,
        new_cell_kwargs={}, new_group_kwargs={}, fuzzy_kwargs={},
        new=None, include_empty_cells=False, verbose=False):
    """
    Build a `FiniteElements`-like object from a :class:`~tramway.tessellation.base.CellStats` object.

    Cells with no (trans-)locations are discarded in addition to those with null or negative label.

    Arguments:

        cells (CellStats): tessellation and partition; must contain (trans-)location data.

        new_cell (callable): cell constructor; default is :class:`Locations` or
            :class:`Translocations` depending on the data in `cells`.

        new_group (callable): constructor for groups of cell; default is :class:`FiniteElements`.

        fuzzy (callable): (trans-)location-cell weighting function; the default for
            translocations considers the initial cell, no weight.

        new_cell_kwargs (dict): keyword arguments for `new_cell`.

        new_group_kwargs (dict): keyword arguments for `new_group`.

        fuzzy_kwargs (dict): keyword arguments for `fuzzy`.

        new (callable): legacy argument; use `new_group` instead.

        include_empty_cells (bool): do not discard cells with no (trans-)locations.

    `fuzzy` takes a :class:`~tramway.tessellation.base.Tessellation` object, a cell index (`int`),
    location coordinates (array-like), cell indices (array-like or callable) and the *get_point*
    function returned by `get_locations` or `get_translocations`.

    If the data are translocations, `fuzzy` takes a pair of arrays as location coordinates (initial
    and final locations respectively) and a pair of arrays or callables as cell indices (initial
    and final cells respectively).

    `fuzzy` returns an array of booleans or values of any scalar type that can be evaluated logically.
    The returned array contains as many elements as input (trans-)locations.

    """
    if new is not None:
        # `new` is for backward compatibility
        warn('`new` is deprecated in favor of `new_group`', DeprecationWarning)
        new_group = new
    if not isinstance(cells, tessellation.CellStats):
        raise TypeError('`cells` is not a `CellStats`')
    if cells.points.size == 0:
        raise ValueError('no data points found')
    if isinstance(cells.cell_index, tuple):
        if len(cells.cell_index[0]) == 0:
            raise ValueError('not any point assigned')

    # format (trans-)locations
    coord_cols, trajectory_col, get_var, get_point = identify_columns(cells.points)
    has_precomputed_deltas = isinstance(coord_cols, tuple)
    if has_precomputed_deltas:
        _nan = np.any(np.isnan(np.asarray(get_var(cells.points, coord_cols[0]))))
    else:
        _nan = np.any(np.isnan(np.asarray(cells.points)))
    if _nan:
        raise ValueError('NaN in location data')
    precomputed = ()
    if new_cell is None:
        are_tracked_molecules = trajectory_col is not None and isinstance(cells.points, pd.DataFrame)
        are_translocations = trajectory_col is not None or has_precomputed_deltas
    else:
        if issubclass(new_cell, Locations):
            are_translocations = are_tracked_molecules = False
        elif issubclass(new_cell, Translocations):
            are_translocations = True
            are_tracked_molecules = issubclass(new_cell, TrackedMolecules)
        else:
            raise TypeError('`new_cell` is neither `Locations` nor `Translocations`')
    if are_translocations:
        precomputed = (coord_cols, trajectory_col, get_var, get_point)
    else:
        precomputed = (coord_cols, get_var, get_point)
    if are_translocations:
        initial_point, final_point, initial_cell, final_cell, get_point = \
            get_translocations(cells.points, cells.cell_index, *precomputed)
        if new_cell is None:
            if are_tracked_molecules:
                new_cell = TrackedMolecules
            else:
                new_cell = Translocations
        fuzzy_args = ((initial_point, final_point), (initial_cell, final_cell), get_point)
    else:
        locations, location_index, get_point = \
            get_locations(cells.points, cells.cell_index, *precomputed)
        fuzzy_args = (locations, location_index, get_point)
        if new_cell is None:
            new_cell = Locations

    # assign/weight (trans-)locations to cells
    if fuzzy is None:
        if are_translocations:
            def f(tessellation, cell, translocations, translocation_cell, get_point):
                initial_point, final_point = translocations
                initial_cell, final_cell = translocation_cell
                if callable(initial_cell):
                    return initial_cell(cell)
                else:
                    return initial_cell == cell
        else:
            def f(tesselation, cell, locations, location_cell, get_point):
                if callable(location_cell):
                    return location_cell(cell)
                else:
                    return location_cell == cell
        fuzzy = f

    # time and space columns in (trans-)locations array
    def _bool(a):
        try:
            return bool(a.size)
        except AttributeError:
            return bool(a)
    if isstructured(cells.points):
        time_col = 't'
    else:
        time_col = cells.points.shape[1] - 1
    if cells.tessellation.scaler is not None and _bool(cells.tessellation.scaler.columns):
        space_cols = cells.tessellation.scaler.columns
    elif isstructured(cells.points):
        space_cols = columns(cells.points)
        if 'n' in space_cols:
            not_space = ['n', time_col]
        else:
            not_space = [time_col]
        if isinstance(space_cols, pd.Index):
            space_cols = space_cols.drop(not_space)
        else:
            space_cols = [ c for c in space_cols if c not in not_space ]
    else:
        if time_col == cells.points.shape[1] - 1:
            space_cols = np.arange(time_col)
        else:
            space_cols = np.ones(cells.points.shape[1], dtype=bool)
            space_cols[time_col] = False
            space_cols, = space_cols.nonzero()
    # remove delta columns
    if has_precomputed_deltas:
        _, delta_cols = coord_cols
        space_cols = [ col for col in space_cols if col not in delta_cols ]

    # pre-select cells
    if include_empty_cells:
        if cells.tessellation.cell_label is None:
            J = np.ones(cells.location_count.shape, dtype=bool)
        else:
            J = 0 < cells.tessellation.cell_label
    else:
        if cells.tessellation.cell_label is None:
            J = 0 < cells.location_count
        else:
            J = np.logical_and(0 < cells.location_count, 0 < cells.tessellation.cell_label)

    # select (with the fuzzy filter) and pre-build cells
    _fuzzy, data, hull = {}, {}, {}
    if are_translocations:
        extra = {}
    for j, ok in enumerate(J): # for each cell
        if not ok:
            continue

        # find (trans-)locations for cell j
        i = fuzzy(cells.tessellation, j, *fuzzy_args, **fuzzy_kwargs)
        if i.dtype in (bool, np.bool8, np.bool_):
            _fuzzy[j] = None
        else:
            _fuzzy[j] = i[i != 0]
            i = i != 0

        if are_translocations:
            _origin = get_point(initial_point, i)
            _destination = get_point(final_point, i)
            if has_precomputed_deltas:
                assert np.all(_origin.index == _destination.index)
                __origin = _origin
            else:
                __origin = _origin.copy() # make copy
                __origin.index += 1
                try:
                    _ok = np.array([i in _destination.index for i in __origin.index])
                    _ok &= np.array([i in __origin.index for i in _destination.index])
                except TypeError:
                    J[j] = False
                    continue
                _origin = get_point(_origin, _ok)
                __origin = get_point(__origin, _ok)
                _destination = get_point(_destination, _ok)
            _points = _origin # for convex hull; ideally not only origins
            points = _destination - __origin # translocations
        else:
            points = _points = get_point(locations, i) # locations

        assert not np.any(np.isnan(np.asarray(points)))

        # convex hull
        _points = np.asarray(get_var(_points, space_cols))
        try:
            hull[j] = cells.tessellation.cell_volume[j]
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            #try:
            #    hull[j] = scipy.spatial.qhull.ConvexHull(_points)
            #except (KeyboardInterrupt, SystemExit):
            #    raise
            #except Exception as e:
            raise
            warn(str(e), RuntimeWarning)

        if not include_empty_cells and points.size == 0:
            J[j] = False
        else:
            if are_tracked_molecules:
                _trajectory_index = cells.points['n'][_origin.index].values
                extra[j] = (_origin, _destination, _trajectory_index)
            elif are_translocations:
                extra[j] = (_origin, _destination)
            data[j] = points

    # simplify the adjacency matrix
    try:
        _adjacency = cells.tessellation.diagonal_adjacency
    except AttributeError:
        _adjacency = None
    _adjacency = cells.tessellation.simplified_adjacency(adjacency=_adjacency, label=J, format='csr')
    ## reweight each row i as 1/n_i where n_i is the degree of cell i
    #n = np.diff(_adjacency.indptr)
    #_adjacency.data = np.repeat(1.0 / np.maximum(1, n), n)

    # build every cells
    J, = np.nonzero(J)
    _cells = OrderedDict()
    for j in J: # for each cell
        try:
            center = cells.tessellation.cell_centers[j]
        except AttributeError:
            center = span = None
        else:
            adj = _adjacency[j].indices
            span = cells.tessellation.cell_centers[adj] - center

        # make cell object
        args = ()
        _volume = None
        try:
            _hull = hull[j]
            _hull.simplices
        except AttributeError:
            _volume = _hull
        except KeyError:
            pass
        else:
            args = (_hull,)
        _cells[j] = new_cell(j, data[j], center, span, *args, **new_cell_kwargs)
        _cells[j].time_col = time_col
        _cells[j].space_cols = space_cols
        if _volume:
            _cells[j].volume = _volume
        if are_translocations:
            try:
                _cells[j].origins = extra[j][0]
            except AttributeError: # `_cells` does not have the `origins` attribute
                pass
            try:
                _cells[j].destinations = extra[j][1]
            except AttributeError: # `_cells` does not have the `destinations` attribute
                pass
            try:
                _cells[j].n = extra[j][2]
            except AttributeError: # `_cells` does not have the `n` attribute
                pass
            except IndexError: # extra[j] has two elements
                pass
        try:
            _cells[j].fuzzy = _fuzzy[j]
        except AttributeError: # `_cells` does not have the `fuzzy` attribute
            pass

    # perform a few last checks
    if verbose:
        if hull:
            for j in J:
                if j not in hull:
                    print('cannot compute convex hull for cell {}'.format(j))
        else:
            print('cannot compute convex hulls')

    #print(sum([ c.tcount == 0 for c in _cells.values() ]))
    self = new_group(_cells, _adjacency, **new_group_kwargs)
    self.tcount = cells.points.shape[0]
    #self.dim = cells.points.shape[1]
    self.ccount = self.adjacency.shape[0]

    return self



class DistributeMerge(FiniteElements):

    __slots__ = ('merged', )

    def __init__(self, cells, adjacency, min_location_count=None):
        if min_location_count is None:
            raise ValueError('`min_location_count` is not defined')
        FiniteElements.__init__(self, cells, adjacency)
        count = np.array([ len(cells[i]) for i in cells ])
        index, = (count < min_location_count).nonzero()
        ordered_index = np.argsort(count[index])
        ordered_index = index[ordered_index]
        index = set(index)
        merged = dict()
        for i in ordered_index[::-1]:
            js = [ j for j in self.neighbours(i).tolist() if j not in index ]
            if not js:
                warn('no large-enough neighbours for cell {:d}'.format(i), RuntimeWarning)
                continue
            dr = np.vstack([ cells[j].center for j in js ]) - cells[i].center#[np.newaxis,:]
            j = js[np.argmin(np.sum(dr * dr, axis=1))]
            j = merged.get(j, j)
            # merge cells i and j
            # attributes: center(?),time_data,space_data,span,boundary,fuzzy,origins,destinations
            cells[j].data = (np.vstack((cells[j].space_data, cells[i].space_data)),
                    np.r_[cells[j].time_data, cells[i].time_data])
            if cells[j].span is not None:
                try:
                    cells[j].span = np.vstack((cells[j].span, cells[i].span))
                except (AttributeError, TypeError):
                    pass
            if cells[j].boundary is not None:
                # cannot merge properly
                try:
                    cells[j].boundary = np.vstack((cells[j].boundary, cells[i].boundary))
                except (AttributeError, TypeError):
                    pass
            try:
                cells[j].fuzzy = np.r_[cells[j].fuzzy, cells[i].fuzzy]
            except (AttributeError, ValueError):
                pass
            try:
                cells[j].origins = np.vstack((cells[j].origins, cells[i].origins))
                cells[j].destinations = np.vstack((cells[j].destinations, cells[i].destinations))
            except AttributeError:
                pass
            #
            merged[i] = j
            index.remove(i)
        # clear and register the merged cells
        discarded = set(ordered_index)
        self.cells = type(cells)([(i, cells[i]) for i in cells if i not in discarded])
        self.merged = dict()
        for i, j in merged.items():
            try:
                self.merged[j].add(i)
            except KeyError:
                self.merged[j] = set([i])
        # modify the adjacency matrix
        adjacency = sparse.lil_matrix(adjacency.shape, dtype=adjacency.dtype)
        for i in range(adjacency.shape[0]):
            if i in discarded:
                continue
            js = set(self.neighbours(i).tolist())
            try:
                ks = self.merged[i]
            except KeyError:
                pass
            else:
                for k in ks:
                    js |= set(self.neighbours(k).tolist())
            js -= discarded
            for j in js:
                adjacency[i,j] = True
                adjacency[j,i] = True
        self.adjacency = adjacency.tocsr()
        assert set(self.cells.keys()) == set((self.adjacency.indptr[1:] - self.adjacency.indptr[:-1]).nonzero()[0].tolist())
        assert set(self.cells.keys()) == set(self.adjacency.indices.tolist())


class Maps(Lazy):
    """
    Basic container for maps, posteriors and the associated input parameters used to generate
    the maps.

    Attributes `distributed_translocations`, `partition_file`, `tessellation_program`, `version`
    are deprecated and will be removed.

    Functional dependencies:

    * setting `maps` unsets `_features`

    """
    __lazy__ = Lazy.__lazy__ + ('_features',)

    def __init__(self, maps, mode=None, posteriors=None):
        Lazy.__init__(self)
        self.maps = maps
        self.mode = mode
        self.min_diffusivity = None
        self.localization_error = None
        self.diffusivity_prior = None
        self.potential_prior = None
        self.jeffreys_prior = None
        self.extra_args = None
        self.distributed_translocations = None # legacy attribute
        self.partition_file = None # legacy attribute
        self.tessellation_param = None # legacy attribute
        self.version = None # legacy attribute
        self.runtime = None
        self.posteriors = posteriors

    @property
    def maps(self):
        return self._maps

    @maps.setter
    def maps(self, m):
        if m is None:
            self._features = None
        else:
            if not isinstance(m, pd.DataFrame):
                raise TypeError('DataFrame expected')
            _err = TypeError('not any column defined')
            if isinstance(m.columns, (tuple, list)):
                if not m.columns:
                    raise _err
            elif m.columns.size == 0:
                raise _err
            self._features = splitcoord(m.columns)
        self._maps = m

    @property
    def features(self):
        """
        `list` or ``None``

        List of different feature names without the coordinate suffix.

        For example, if `maps` has columns *diffusivity*, *force x* and *force y*, `features`
        will return *diffusivity* and *force*.
        """
        if self._features is None:
            return None
        else:
            return list(self._features.keys())

    @property
    def variables(self):
        """
        Alias for `features`.
        """
        warn('`variables` is deprecated; use `features` instead', PendingDeprecationWarning)
        return self.features

    @property
    def _variables(self):
        """
        Alias for `_features`.
        """
        warn('`_variables` is deprecated; use `_features` instead', PendingDeprecationWarning)
        return self._features

    def __nonzero__(self):
        return self.maps is not None

    def __len__(self):
        return 0 if self._features is None else len(self._features)

    def __contains__(self, feature_name):
        return feature_name in self._features

    def __getitem__(self, feature_name):
        try:
            if isinstance(feature_name, str):
                return self.maps[dict(self._features)[feature_name]]
            else:
                import itertools
                return self.maps[list(itertools.chain(*[
                            dict(self._features)[ftr] for ftr in feature_name
                        ]))]
        except (TypeError, KeyError):
            raise KeyError("no such mapped feature: '{}'".format(feature_name))

    def __setitem__(self, feature_name, val):
        if isinstance(val, np.ndarray):
            if val.size == self.maps.shape[0]:
                self.maps[feature_name] = val
                self._features = None
            else:
                raise ValueError('wrong array size')
        else:
            raise NotImplementedError('only numpy ndarrays are supported')

    def sub(self, ix, reindex=False):
        """
        Sub-map.

        Operates like `loc` on a `DataFrame`.

        Arguments:

            ix (slice or array-like):
                cell indices or boolean array.

            reindex (bool):
                make index range from 0 without missing integer.

        Returns:

            tramway.inference.base.Maps: sub-map.

        """
        sub_map = copy(self)
        if isinstance(ix, slice):
            ix = np.arange(ix.start, ix.stop, ix.step)
        sub_map.maps = self.maps.loc[ix]
        if reindex:
            sub_map.maps.index = np.arange(sub_map.maps.index.size)
        return sub_map

    def __str__(self):
        attrs = { k: v for k, v in self.__dict__.items() if not (k[0] == '_' or v is None) }
        v = self.features
        if v is not None:
            attrs['features'] = v
        attrs['maps'] = type(self.maps)
        l = max(len(k) for k in attrs)
        s = '\n'.join([ '{}:{} {}'.format(k, ' '*(l-len(k)), str(v)) for k, v in attrs.items() ])
        return s

    def defattr(self, attr, val):
        count = 0
        while True:
            try:
                getattr(self, attr)
            except AttributeError:
                break
            else:
                if count:
                    attr = attr[:-1] + str(count)
                else:
                    if attr[-1] in '0123456789' or '_' in attr:
                        attr = attr + '_0'
                    else:
                        attr = attr + '0'
                count += 1
        setattr(self, attr, val)



class OptimizationWarning(RuntimeWarning):
    pass


class DiffusivityWarning(RuntimeWarning):
    def __init__(self, diffusivity, lower_bound=None):
        self.diffusivity = diffusivity
        self.lower_bound = lower_bound

    def __repr__(self):
        if self.lower_bound is None:
            return 'DiffusivityWarning({})'.format(self.diffusivity)
        else:
            return 'DiffusivityWarning({}, {})'.format(self.diffusivity, self.lower_bound)

    def __str__(self):
        if self.lower_bound is None:
            return 'wrong diffusivity value: {}'.format(self.diffusivity)
        else:
            return 'diffusivity too low: {} < {}'.format(self.diffusivity, self.lower_bound)


def smooth_infer_init(cells, min_diffusivity=None, jeffreys_prior=None, **kwargs):
    """
    Initialize the diffusivity array for translocations.

    This function is a collection of checks and initial computations suitable for *infer* functions
    in inference plugins.

    Arguments:

        cells (FiniteElements): first input argument of the *infer* function.

        min_diffusivity (float): minimum allowed diffusivity value.

        jeffreys_prior (bool): activate Jeffreys' prior.

    Returns:

    * *index* (:class:`numpy.ndarray`) -- cell indices corresponding to arrays *n*, *dt_mean* and *D_initial*.
    * *reverse_index* (:class:`numpy.ndarray`) -- element indices in arrays *index*, *n*, *dt_mean*, *D_initial* for each cell index.
    * *n* (:class:`numpy.ndarray`) -- translocation count for each cell.
    * *dt_mean* (:class:`numpy.ndarray`) -- mean translocation duration for each cell.
    * *D_initial* (:class:`numpy.ndarray`) -- initial diffusivity array.
    * *min_diffusivity* (:class:`float` or ``None``) -- minimum diffusivity value; the returned value depends on `jeffreys_prior`.
    * *D_bounds* (:class:`list`) -- list of (lower bound, upper bound) couples; suitable as *bounds* input argument for the :func:`scipy.optimize.minimize` function.
    * *border* (:class:`numpy.ndarray`) -- MxD boolean matrix with M the number of cells and D the number of space dimensions

    """
    # initial values and sanity checks
    index, n, dt_max, dt_mean, D_initial, border = [], [], [], [], [], []
    reverse_index = np.full(cells.adjacency.shape[0], -1, dtype=int)

    j = 0
    for i in cells:
        cell = cells[i]

        #assert i == cell.index # NO!
        if not bool(cell):
            raise ValueError('empty cells')

        # sanity checks
        if cell.dr.shape[1] == 0:
            raise ValueError('translocation array has no column')
        if cell.dt.shape[1:]:
            raise ValueError('time deltas are structured in multiple dimensions')
        if np.any(np.isnan(cell.dt)):
            raise ValueError('time delta is nan')
        # ensure that translocations are properly oriented in time
        if not np.all(0 < cell.dt):
            warn('translocation dts are not all positive', RuntimeWarning)
            cell.dr[cell.dt < 0] *= -1.
            cell.dt[cell.dt < 0] *= -1.

        # check cell i has neighbours
        try:
            adjacent = cells.adjacency.indices[cells.adjacency.indptr[i]:cells.adjacency.indptr[i+1]]
            adjacent = [ c for c in adjacent if cells[c] ]
            if not adjacent:
                continue
        except ValueError:
            continue

        # initialize the local diffusivity parameter
        dt_max_i = np.max(cell.dt)
        dt_mean_i = np.mean(cell.dt)
        D_initial_i = np.mean(cell.dr * cell.dr) / (2. * dt_mean_i)
        #

        index.append(i)
        reverse_index[i] = j
        j += 1
        n.append(float(len(cell)))
        dt_max.append(dt_max_i)
        dt_mean.append(dt_mean_i)
        D_initial.append(D_initial_i)

        # border
        try:
            if cell.center is None:
                warn('missing cell center', RuntimeWarning)
                border.append(np.zeros(cell.dim, dtype=np.bool_))
            else:
                adjacent = np.vstack([ cells[c].center for c in adjacent if cells[c] ])
                border.append(np.logical_or(
                    np.max(adjacent, axis=0) <= cell.center,
                    cell.center <= np.min(adjacent, axis=0)
                    )) # to be improved
        except ValueError:
            border.append(None)

    n, dt_mean, D_initial = np.array(n), np.array(dt_mean), np.array(D_initial)

    if min_diffusivity is None:
        noise_dt = kwargs['sigma2']
        D_bounds = [( (1e-16 -noise_dt)/dt_max_i, None ) for dt_max_i in dt_max ]
        min_diffusivity = 0
    else:
        if min_diffusivity is False:
            min_diffusivity = None
        D_bounds = [(min_diffusivity, None)] * D_initial.size

    try:
        border = np.vstack(border)
    except:
        border = None

    return index, reverse_index, n, dt_mean, D_initial, min_diffusivity, D_bounds, border


__all__ = ['Local', 'Distributed', 'Cell', 'Locations', 'Translocations', 'Maps',
    'FiniteElement', 'FiniteElements',
    'identify_columns', 'get_locations', 'get_translocations', 'distributed',
    'TrackedMolecules', 'DistributeMerge',
    'DiffusivityWarning', 'OptimizationWarning', 'smooth_infer_init']





from . import *
from ..attribute import initializer_method

from_voronoi                = initializer_method( SamplerInitializer.from_voronoi                )
from_spheres                = initializer_method( SamplerInitializer.from_spheres                )
from_nearest_neighbors      = initializer_method( SamplerInitializer.from_nearest_neighbors      )
from_nearest_time_neighbors = initializer_method( SamplerInitializer.from_nearest_time_neighbors )




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.tessellation.base import Partition
from tramway.tessellation.time import TimeLattice
from tramway.plot.map import *
from tramway.plot.animation import *
import numpy as np


def animate_map_2d(_map, cells, output_file=None,
        frame_rate=1, bit_rate=None, dots_per_inch=200, play=False,
        time_step=None, time_unit='s', figure=None, axes=None,
        bounding_box=None, colormap=None, colorbar=True, axis=True,
        verbose=True, time_precision=2, **kwargs):
    """
    Animate 2D maps.

    Arguments:

        _map (pandas.DataFrame): scalar or vector map.

        cells (tramway.tessellation.base.Partition): spatial-temporal binning.

        output_file (str): path to .mp4 file.

        frame_rate (float): number of frames per second.

        bit_rate (int): movie bitrate; see also :class:`~matplotlib.animation.FFMpegWriter`.

        dots_per_inch (int): dpi argument passed to :meth:`~matplotlib.animation.FFMpegWriter.saving`.

        play (bool): play the movie once generated.

        time_step (float): time step between successive frames.

        time_unit (str): time unit for time display; if None, no title is displayed.

        figure (matplotlib.figure.Figure): figure handle.

        axes (matplotlib.axes.Axes): axes handle.

        bounding_box (matplotlib.transforms.Bbox): axis bounding box.

        colormap (str): colormap name; see also https://matplotlib.org/users/colormaps.html.

        colorbar (bool or str or dict): see also :func:`~tramway.plot.map.scalar_map_2d`.

        axis (bool or str): make the axes visible.

        verbose (bool): ask for confirmation if the output file is to be overwritten
            and draw a progress bar (module :mod:`tqdm` must be available).

        time_precision (int): number of decimals for time display.

    Extra keyword arguments are passed to :class:`~matplotlib.animation.FFMpegWriter`.

    """

    dim = _map.shape[1]
    if dim == 1:
        clim = [_map.values.min(), _map.values.max()]
        _render = scalar_map_2d
    elif dim == 2:
        _amplitude = _map.pow(2).sum(1).apply(np.sqrt)
        clim = [_amplitude.values.min(), _amplitude.values.max()]
        _render = field_map_2d
    else:
        raise ValueError('nD data not supported for n not 1 or 2')

    if isinstance(cells.tessellation, TimeLattice) \
            and cells.tessellation.spatial_mesh is not None:
        segments = cells.tessellation.time_lattice
        tmin, tmax = segments.min(), segments.max()
        _map = cells.tessellation.split_frames(_map)
        _mesh = cells.tessellation.spatial_mesh
        cells = Partition(tessellation=_mesh, location_count=np.ones(_mesh.number_of_cells), points=cells.points)
    else:
        _map = [_map]
        t = cells.locations['t']
        tmin, tmax = t.min(), t.max()
        segments = [[tmin, tmax]]

    if time_step is None:
        N = len(segments) # len(_map)
    else:
        N = round((tmax - tmin) / time_step) + 1
        if frame_rate is None:
            frame_rate = 1. / time_step # assume dt is in seconds

    if time_unit:
        title_pattern = "time = {{:.{:d}f}} {}".format(time_precision, time_unit)

    try:
        with VideoWriterReader(output_file, frame_rate, bit_rate, dots_per_inch,
                figure, axes, axis, verbose, **kwargs) as movie:

            with movie.saving():
                for f in movie.range(N):
                    if time_step is None:
                        t = np.mean(segments[f])
                        __map = _map[f]
                    else:
                        t = tmin + f * time_step

                        seg_scale = np.median([ _s[1] - _s[0] for _s in segment ]) * .5
                        segs = []
                        weights = []
                        for s, seg in enumerate(segments):
                            if seg[0] <= t and t <= seg[1]:
                                segs.append(s)
                                seg_center = (seg[0] + seg[1]) / 2.
                                w = 1. - np.abs(t - seg_center) / seg_scale
                                weights.append(w)
                        assert bool(segs)

                        weights = np.array(weights)
                        if np.all(weights == 0):
                            weights[...] = 1. / numel(weights)
                        else:
                            weights /= np.sum(weights)
                        weights = list(weights)

                        __map = _map[segs.pop()] * weights.pop()
                        for s, w in zip(segs, weights):
                            if w == 0:
                                continue
                            __map = __map + _map[s] * w

                    _render(cells, __map, clim=clim, figure=movie.figure, axes=movie.axes,
                            colorbar=colorbar and f==0, colormap=colormap)
                    if bounding_box is not None:
                        movie.axes.update_datalim_bounds(bounding_box)
                    #
                    if time_unit:
                        movie.axes.set_title(title_pattern.format(t))
                    #
                    movie.grab_frame()
                    movie.axes.clear()

            if play:
                movie.play()

    except Aborted:
        pass


__all__ = ['animate_map_2d']

